import brian_no_units
from brian.units import *
from brian.group import *
import time
import scipy.signal as sg
import scipy.stats as stats
import numpy as np
from brian.stdunits import *
from brian.network import run, Network, network_operation
from brian.connections.connection import Connection
from brian.directcontrol import SpikeGeneratorGroup
from brian.clock import defaultclock
from brian.equations import Equations
from brian.neurongroup import NeuronGroup
from brian.monitor import SpikeMonitor, StateMonitor


defaultclock.dt = 0.0001*second
#Network Parameters:
N_EP1=124;
N_IP1=4;
N = N_EP1+N_IP1
#VLSI Neuron constants
sigma_mismatch=0.25
    
def create_netobjs(stim, params):
    C         = params['Cap']   
    KappaN    = params['Kappan']
    KappaP    = params['Kappap']
    I0N       = params['I0n']
    I0P       = params['I0p']
    Ut        = params['Ut']
    Delta_t   = (Ut/KappaP)
    
    #Feed-Forward parameters
    i_inj           =       params['i_inj'    ]
    i_injinh1       =       params['i_injinh1']
    i_injinh2       =       params['i_injinh2']
    i_injinh3       =       params['i_injinh3']
    i_injinh4       =       params['i_injinh4']
    i_leak          =       params['i_leak'   ]
    tau_syn_E       =       params['tau_syn_E']
    tau_syn_I       =       params['tau_syn_I']
    tau_synloc_E    =       params['tau_synloc_E']
    tau_synloc_IE   =       params['tau_synloc_IE']
    v_thresh        =       params['v_thresh' ]
    w_syn_E1        =       params['w_syn_E1' ]
    w_syn_E2        =       params['w_syn_E2' ]
    w_syn_I         =       params['w_syn_I'  ]
    w_synloc_E      =       params['w_synloc_E']
    w_synloc_EE     =       params['w_synloc_EE']
    w_synloc_EI     =       params['w_synloc_EI']
    w_synloc_IE     =       params['w_synloc_IE']
    w_synloc_S      =       params['w_synloc_S']
    ##Feed-Back parameters
    #w        =dict()
    #w['e']    =1e-11
    #w['i']    =5e-11/N_IP1
    #w['ei']    =2e-11/N_EP1
        
    eqs    =Equations('''
    dV/dt=(-I_lk + I_fb + I_in + Ia + Iloce - Iloci - Ii)/C: volt
    
    
    I_fb = I0P*exp((V-v_thresh)/Delta_t) : amp
    I_in : amp
    I_lk = i_leak : amp
    
    dIa/dt=-Ia/tau_syn_E: amp
    dIloce/dt=-Iloce/tau_synloc_E: amp
    dIloci/dt=-Iloci/tau_synloc_IE: amp
    dIi/dt=-Ii/tau_syn_I: amp      
    ''')
        
    EIP    = NeuronGroup(N, model=eqs, reset = 0, threshold=1.5, refractory=.001)
    EP1    = EIP[:N_EP1]
    IP1    = EIP[N_EP1:]
    
    EP1.I_in = np.random.normal(1,sigma_mismatch,N_EP1)*i_inj
    IP1.I_in = np.random.normal(1,sigma_mismatch,N_IP1)*\
                                   np.array([i_injinh1,
                                             i_injinh2,
                                             i_injinh3,
                                             i_injinh4])
    
    #Create connections between population
    v_loclat = np.zeros([N_EP1])
    v_loclat[N_EP1/2] = w_synloc_S
    v_loclat[[N_EP1/2-1,N_EP1/2+1]] = w_synloc_E
    v_loclat[[N_EP1/2-2,N_EP1/2+2]] = w_synloc_EE
    v_loclat[[N_EP1/2-3,N_EP1/2+3]] = w_synloc_EE/2
    v_loclat = np.roll(v_loclat,-N_EP1/2)
    W = np.array([ np.roll(v_loclat,i) for i in range(N_EP1)])
    W *= np.random.normal(1,sigma_mismatch,W.shape)    
    ConnE    =Connection(EP1,EP1,'Iloce');ConnE.connect(EP1,EP1,W)
    ConnEI   =Connection(EP1,IP1,'Iloce');ConnEI.connect(EP1,IP1, W = w_synloc_EI*np.random.normal(1,sigma_mismatch,[len(EP1),len(IP1)]))
    ConnIE   =Connection(IP1,EP1,'Iloci');ConnIE.connect(IP1,EP1, W = w_synloc_IE*np.random.normal(1,sigma_mismatch,[len(IP1),len(EP1)]))
    
    M_EIP =SpikeMonitor(EIP)    
    MV_EIP= StateMonitor(EIP,'V',record=list(range(0,N)),timestep=int(1*ms/defaultclock.dt))

    @network_operation
    def update_mpot():
        EIP.V[EIP.V<0.]=0.
#     ME_EP1= StateMonitor(EP1,'Ie',record=range(0,N_EP1),timestep=int(1*ms/defaultclock.dt))
#     MI_EP1= StateMonitor(EP1,'Ii',record=range(0,N_EP1),timestep=int(1*ms/defaultclock.dt))
#     MW_EP1= StateMonitor(EP1,'Ia',record=range(0,N_EP1),timestep=int(1*ms/defaultclock.dt))
    
    netobjs =     {'EIP':EIP,                
                   'update_mpot':update_mpot,                                    
                   'ConnE': ConnE,
                   'ConnEI': ConnEI,
                   'ConnIE': ConnIE,
                   #'M_In1': M_In1,
                   'M_EIP': M_EIP,
                   'MV_EIP': MV_EIP}
                   
    return netobjs, M_EIP, MV_EIP
    


if __name__ == '__main__':
    from pylab import *
    from brian.plotting import raster_plot
    ion()
    from .paramTranslation import params, loadBiases
    from expSetup import *
    configurator = pyNCS.ConfAPI.Configurator()
    configurator._readCSV('chipfiles/ifslwta.csv')
    configurator.set_parameters(loadBiases('biases/defaultBiases_ifslwta'))
    p=params(configurator, 'chipfiles/ifslwta_paramtrans.xml')
    p.translate('pinj',2.8)    
    p.translate('nsynloclat1',0.55)
    p.translate('nsynloclat2',0.53)
    p.translate('nsynexcinh',0.45)   
    p.translate('psynlocinhw',2.75)
    p.translate('psynlocinhth',.3)
    stim = np.transpose([np.random.randint(0,128,32000), np.cumsum(np.random.random(32000)/16)])
    netobjs,  M_EIP, MV_EIP  = create_netobjs(stim,p.cur)
    net = Network(list(netobjs.values()))
    net.run(1)
    raster_plot(*[M_EIP])
    





#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Emre Neftci
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

from .STas import events, channelEvents, RawOutput, channelAddressing, addrSpec
from .STas import setDefaultMonChannelAddress, setDefaultSeqChannelAddress,\
                 getDefaultMonChannelAddress, getDefaultSeqChannelAddress
from .STsl import STCreate
from .spikes import SpikeList, SpikeTrain, merge, merge_spikelists, \
                   merge_sequencers
from . import pyST_globals
import numpy as np


def loadtxt(filename, comments, format='t'):
    data = np.loadtxt(filename, comments=comments)
    if format == 't':
        data = np.fliplr(data)
    ev = events(data, atype='l')
    id_list = np.unique(ev.get_ad())
    return SpikeList(ev.get_adtmev(), id_list)



#!/bin/python
#-----------------------------------------------------------------------------
# File Name : testComAPI.py
# Purpose: Brian Simulator API for testing pyNCS. Can only be used as one 
# instance because mappings are global 
#
# Author: Emre Neftci
#
# Creation Date : 05-06-2013
# Last Modified : Thu 26 Mar 2015 12:21:07 PM PDT
#
# Copyright : (c) 
# Licence : GPLv2
#----------------------------------------------------------------------------- 
from pyNCS.neurosetup import get_data
from pyNCS.api.ComAPI import *
from pyNCS.api.ConfAPI import *
import pyNCS.pyST as pyST
from . import IFSLWTA_brian_model as ibm
import numpy as np
from brian.directcontrol import PoissonGroup
import os

#TODO: multichip mappings
#TODO: generalize _decode_mappings_list

chipname = 'ifslwta'
channel_seq = 0
channel_ifslwtain = 6
global_mappings_list = []
global_neurosetup = None
global_sympy_params = None

synapse_trans_dict = {0 : ['Ii', 'w_syn_I'],
                      1 : ['Ii', 'w_syn_I'],
                      2 : ['Ia', 'w_syn_E1'],
                      3 : ['Ia', 'w_syn_E2']}

def _decode_mappings_list(mappings_list, synapse_id):
    ml = np.array(mappings_list)
    fr = decode_addr(ml[:,0])[channel_seq] #0 for field 0
    to = decode_addr(ml[:,1])[channel_ifslwtain]
    to_syn = to[0,to[1,:]==synapse_id]
    fr_syn = fr[0,to[1,:]==synapse_id]
    if len(to_syn)==0:
        return []
    else:        
        if np.shape(ml)[1]==3:
            prob = ml[:,2].astype('float')/128
            prob_syn = prob[to[1,:]==synapse_id]
        else:    
            prob = np.ones(ml.shape[0])            
        return list(zip(fr_syn,to_syn,prob_syn))

def _dlist_to_mappingdict(mapping):
    from collections import defaultdict
    #sort list
    if len(mapping)>0:
        mapping_dict = defaultdict(list)
        mapping_dict_probs = defaultdict(list)
        def func(srctgt):
            mapping_dict[srctgt[0]].append(srctgt[1])
            mapping_dict_probs[srctgt[0]].append(srctgt[2])
        list(map(func, mapping))
        return mapping_dict, mapping_dict_probs
    else:
        return {},{}

def _mappingdict_to_matrix(mapping_dict, mapping_dict_probs):
    N_fr = len(mapping_dict)
    N_to = max(max(mapping_dict.values()))+1
    M = np.zeros([N_fr, N_to])
    P = np.zeros([N_fr, N_to])
    for k in list(mapping_dict.keys()):
        M[k,mapping_dict[k]] = 1
        P[k,mapping_dict[k]] = mapping_dict_probs[k] 
    return M, P

def translate_mappings(mapping_list, synapse_id):
    m = _decode_mappings_list(mapping_list, synapse_id)
    if m ==None:
        return None
    else:
        return _mappingdict_to_matrix(*_dlist_to_mappingdict(m))

def decode_addr(addr):
    #decode physical addresses to neuron - synapse    
    global global_neurosetup    
    return global_neurosetup.seq.addrPhysicalExtract(addr)    

class Communicator(BatchCommunicatorBase):
    def run(self, stimulus=None, duration=None, context_manager=None):
        stimulus_abs = pyST.events(stimulus, atype='p', isISI=True)
        stimulus_abs.set_abs_tm()
        self.evn_sin = evs_in = stimulus_abs.get_adtmev().astype('float')
        evs_in[:,1]*=1e-6 #brian takes seconds
        if duration == None and len(stimulus_abs)>0:
            duration = np.max(evs_in[:,1])
            print(duration)
        elif duration == None:
            duration = 1.
        else:
            duration = float(duration)/1e3
        net, M_EIP, MV_EIP = self._prepare_brian_net(evs_in)
        self.outs = [net, M_EIP, MV_EIP]
        net.reinit(states=False)
        print(('running virtual IFLSWTA for {0}s'.format(duration)))  
        net.run(duration)
        sp = np.array(M_EIP.spikes).reshape(-1,2)
        sp[:,0]+=(2**16) #slot 2 in text.xml. THIS IS A POSSIBLE SOURCE OF TEST ERROR
        sp[:,1]*=1e6 
        return sp
        
    
    def _prepare_brian_net(self, stimulus):
        from brian.directcontrol import SpikeGeneratorGroup
        from brian.synapses.synapses import Synapses
        from brian.network import Network
        self.S = []
        netobjs, M_EIP, MV_EIP = ibm.create_netobjs(stimulus, global_sympy_params.cur)
        
        if len(stimulus)>0: 
            N = max(stimulus[:,0])+1
            In1=SpikeGeneratorGroup(N,stimulus[stimulus[:,0] <N])
            self.S.append(In1)
        #Create connections, if any
        if len(global_mappings_list)>0:                 
            gml = np.array(global_mappings_list).reshape(-1,3)
            for syn_idx in range(4):   
                      
                gml_dict, pgml_dict = _dlist_to_mappingdict(_decode_mappings_list(gml,syn_idx))
                if len(gml_dict)>0:                
                    if len(stimulus)==0:
                        input_pop = PoissonGroup(max(gml_dict.keys())+1, 0)
                    else:
                        input_pop = In1
                    iname, wname = synapse_trans_dict[syn_idx] 
                    EIP = netobjs['EIP']
                    S=Synapses(input_pop,EIP,model="""w : 1
                                                p : 1""",
                                     pre=iname+"+=w*(rand()<p)")
                     
                    for i in gml_dict:
                        S[i,gml_dict[i]]=True                
                        S.w[i,gml_dict[i]] = global_sympy_params.cur[wname]*np.random.normal(1,ibm.sigma_mismatch, len(gml_dict[i]))
                        S.p[i,gml_dict[i]] = pgml_dict[i] 
                                    
                    self.S.append(S)
            
        net = Network(list(netobjs.values())+self.S)
                        
        return net, M_EIP, MV_EIP  
        
    
class Configurator(ConfiguratorBase):        
    def register_neurosetup(self, neurosetup):
        '''
        Provides a link to the Neurosetup. This is useful for complex parameter
        configuration protocols requiring the sequencing and monitoring of
        address-events
        '''
        self._neurosetup_registered = True
        self._neurosetup = neurosetup
        global global_neurosetup
        global_neurosetup = neurosetup
        
    def get_parameter(self, param_name):
        #IMPLEMENT
        '''
        Gets parameter param_name.
        '''
        return self.parameters[param_name]        
    
    def set_parameter(self, param_name, param_value):
        #IMPLEMENT
        '''
        Sets parameter param_name with param_value
        '''
        param_dict = {param_name:param_value}
        self.sympy_params.cur.update(param_dict)   
        self.sympy_params.update()    
        self.parameters.update(param_dict)
        return None
    
    def set_parameters(self, param_dict):
        #IMPLEMENT
        self.sympy_params.cur.update(param_dict)  
        self.sympy_params.update()     
        self.parameters.update(param_dict)
        return None
    
    
    def add_parameter(self, param):
        #CONVENIENCE FUNCITON. IMPLEMENTATION NOT REQUIRED
        '''
        Add a parameter to the configurator
        param: dictionary with all attributes of parameter or an xml element or
        file name with the <parameter /> element
        '''
        if isinstance(param, dict):            
            self.parameters[param['SignalName']] = 0
        elif isinstance(param, etree._Element):
            self.parameters[param.SignalName] = 0
    
    def reset(self):
        #IMPLEMENT
        '''
        Resets all the parameters to default values
        '''
        return None
    
    def __parseNHML__(self, doc):
        '''
        Parse xml file or element tree to generate the object
        '''
        super(Configurator,self).__parseNHML__(doc)
        global global_sympy_params
        from .paramTranslation import params
        filename = get_data('chipfiles/ifslwta_paramtrans.xml')
        global_sympy_params = self.sympy_params = params(self, filename)
        
    def _readCSV(self, CSVfile):        
        super(Configurator,self)._readCSV(CSVfile)
        global global_sympy_params
        from .paramTranslation import params
        filename = get_data('chipfiles/ifslwta_paramtrans.xml')
        global_sympy_params = self.sympy_params = params(self, filename)
    
class Mappings(MappingsBase):
    def add_mappings(self, mappings):
        #IMPLEMENT (REQUIRED)
        '''
        Adds *mappings* to the mappings table.

        Inputs:
        *mappings*: a two-dimenstional iterable
        '''
        global global_mappings_list
        global_mappings_list += mappings 

    def get_mappings(self):
        #IMPLEMENT (REQUIRED)
        '''
        Returns an array representing the mappings
        '''
        global global_mappings_list        
        return global_mappings_list

    def clear_mappings(self):
        #IMPLEMENT (REQUIRED)
        '''
        Clears the mapping table. No inputs
        '''
        global global_mappings_list
        while True:
            try:
                global_mappings_list.pop()
            except IndexError:
                break
            
        return None




#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Emre Neftci
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------
# -*- coding: utf-8 -*-
######################################################################
# Emre Neftci Author <emre.neftci@ini.phys.ethz.ch>
#######################################################################

import numpy as np
import os
import copy
import time
import warnings
from . STsl import *
import itertools
from contextlib import contextmanager
from . import pyST_globals
from lxml import etree


PERFORMANCE_DEBUG=False

#TODO: RawOutput should return spikelists with complete id_list even when empy
class RawOutput(object):
    '''
    RawOutPut is a class which contains raw AER data per channel (Physical addresses).

    It decodes the AER data "just in time"

    Inputs:
    *raw_data*: events with physical addresses
    *id_data*: all addresses in given channel
    *decoder_dict*: Dictionary of functions for decoding addresses. Key designates the channel
    *t_start*: start time of all spikelists in ms
    *t_stop*: end time of all spikelists in ms
    *filter_duplicates*: If True erases in all channels double events within the a 0.01ms time-frame (buggy hw)

    Usage:
    #In conjunction with pyAex.netClient

    >>> client = client.netClient(host='localhost')
    >>> raw_out = client.stimulate()
    >>> client.stop()
    >>> for sl in raw_output: sl.raster_plot()
    '''
    def __init__(self, raw_data, decoder_dict, t_start=0, t_stop=1000, filter_duplicates=False):

        #Inputs
        self.raw_data = raw_data.copy()
        self.decoder_dict = decoder_dict
        self.filter_duplicates = filter_duplicates

        #Containers
        self.decoded_data = {}
        self.channels = list(self.raw_data.keys())
        self.t_start = t_start
        self.t_stop = t_stop
        self.print_num_events()

    def print_num_events(self):
        lev = []
        for k, v in list(self.raw_data.items()):
            lev.append('Ch{0}: {1} evs '.format(k, len(v)))
        print((''.join(lev)))

    def __getitem__(self, key):
        key = int(key)
        if not key in self.decoded_data:
            self.decode_data(key)

        return self.decoded_data[key]

    def __getstate__(self):
        self.decode_all_channels()
        dc = self.__dict__.copy()
        del dc['decoder_dict']
        return dc

    def decode_all_channels(self):
        for i in self.channels:
            self.decode_data(i)

    def decode_data(self, key):
        with self.check_has_key_somewhere(key):
            if PERFORMANCE_DEBUG: t0=time.time()
            ad_data = self.decoder_dict[key](self.raw_data[key].get_ad())
            tm_data = self.raw_data[key].get_tm().astype('float') / 1000
            if PERFORMANCE_DEBUG:
                print(('Decoding events took {0} seconds'.format(time.time()-t0)))

            if PERFORMANCE_DEBUG: t0=time.time()
            evs = events(atype='l')
            evs.add_adtm(ad_data, tm_data)
            if PERFORMANCE_DEBUG:
                print(('Building events took {0} seconds'.format(time.time()-t0)))

            if PERFORMANCE_DEBUG: t0=time.time()
            st_data = SpikeList( evs.get_adtmev(), np.unique(evs.get_ad()))
            if PERFORMANCE_DEBUG:
                print(('Building SpikeList took {0} seconds'.format(time.time()-t0)))

            if self.filter_duplicates:
                st_data.filter_duplicates()
            self.raw_data.pop(key)
            self.decoded_data[key] = st_data

    def iteritems(self):
        for key, value in self.iterchannels(self.channels):
            yield key, value

    def __iter__(self):
        for key, value in self.iterchannels(self.channels):
            yield value

    def iterchannels(self, channels):
        for key in channels:
            #key is a channel
            #self[key] is a spikelist
            yield key, self[key]

    @contextmanager
    def check_has_key_somewhere(self, key):
        try:
            yield
        except KeyError:            
            if not key in self.decoder_dict:
                raise KeyError("There is no function to decode %d" % key)
            if key in self.decoded_data:
                print(("Channel {0} has already been decoded".format(key)))
            elif not key in self.raw_data:
                #print("Channel {0} is not present, assuming no events".format(key
                self.decoded_data[key] = SpikeList([], [])
            else:
                raise


class events(object):
    #Number of dimensions (address, timestamp)
    NF = 2

    def __init__(self, ev=None, atype='p', isISI=False):
        self.isISI = isISI
        assert isinstance(atype, str)
        __atype = atype[0].lower()

        if __atype == 'p':
            __dtype = np.dtype([('tm', 'uint32'),
                              ('ad', 'uint32')])

        if __atype == 'l':
            __dtype = np.dtype([('tm', 'float'),
                              ('ad', 'float')])

        self.__atype = __atype
        self.__dtype = __dtype

        if isinstance(ev, events):
            self.__data = ev.__data.copy()
        elif ev is not None:
            ev = np.array(ev)
            if ev.shape[1] == self.NF:
                self.set_data(ev[:, 0], ev[:, 1])
            elif ev.shape[0] == self.NF:
                self.set_data(ev[0, :], ev[1, :])
            else:
                raise TypeError
        else:
            self.__data = np.zeros([0], self.dtype)

    @property
    def data(self):
        return self.__data

    @property
    def atype(self):
        return self.__atype

    @property
    def dtype(self):
        return self.__dtype

    def __len__(self):
        return self.get_nev()

    def __add__(self, other):
        self.add_adtm(other.ad, other.tm)

    def __repr__(self):
        return self.get_adtm().__repr__()

    def get_nev(self):
        return len(self.get_tm())

    @property
    def nev(self):
        return self.get_nev()

    def get_tdur(self):
        if self.nev == 0:
            return 0
        if self.isISI:
            return self.tm.sum()
        else:
            if self.get_nev() > 0:
                return self.tm[-1] - self.tm[0]

    @property
    def tdur(self):
        return self.get_tdur()

    def get_t_stop(self):
        if self.isISI:
            return self.tm.sum()
        else:
            if self.get_nev() > 0:
                return self.tm[-1]

    @property
    def t_stop(self):
        return self.get_t_stop()

    def get_ad(self):
        return self.__data['ad']

    def set_ad(self, ad):
        self.__data['ad'] = ad
        return self.get_ad()

    @property
    def ad(self):
        return self.get_ad()

    #@ad.setter
    #def ad(self, value):
    #    return self.set_ad(value)

    def get_tm(self):
        return self.__data['tm']

    def set_tm(self, tm):
        self.__data['tm'] = tm
        return self.get_tm()

    @property
    def tm(self):
        return self.get_tm()

    #@tm.setter
    #def tm(self, value):
    #    return self.set_tm(value)

    def set_data(self, ad, tm):
        assert len(ad) == len(tm), "addresses and timestamps lengths are incompatible %d %d" % (len(ad), len(tm))
        self.__data = np.zeros(len(ad), self.dtype)
        self.__data['ad'] = ad.astype(self.dtype['ad'])
        self.__data['tm'] = tm.astype(self.dtype['tm'])

    def add_adtmev(self, ev):
        if not isinstance(ev, np.ndarray):
            ev = np.array(ev)

        if len(ev.shape) != self.NF:
            ev = ev.reshape(-1, 2).astype(self.dtype)

        ad = np.concatenate([self.ad, ev[:, 0]])
        tm = np.concatenate([self.tm, ev[:, 1]])
        self.set_data(ad, tm)

    def add_adtm(self, ad, tm):
        if not isinstance(ad, np.ndarray):
            ad = np.array(ad)

        if len(ad.shape) != 1:
            ad = ad.reshape(-1, 1)

        if not isinstance(tm, np.ndarray):
            tm = np.array(tm)

        if len(tm.shape) != 1:
            tm = tm.reshape(-1, 1)

        assert tm.shape == ad.shape

        ad = np.concatenate([self.ad, ad])
        tm = np.concatenate([self.tm, tm])
        self.set_data(ad, tm)

    def get_tmad(self):
        return np.array([self.tm, self.ad])

    def get_adtm(self):
        return np.array([self.ad, self.tm])

    def get_tmadev(self):
        return self.get_tmad().transpose()

    def get_adtmev(self):
        return self.get_adtm().transpose()

    def normalize_tm(self, t0=0.):
        if not self.isISI:
            t_start = self.get_tm().min()
            self.set_tm(self.get_tm() - t_start + t0)
        else:
            t = self.get_tm()
            t[0] = t0
            self.set_tm(t)

    def get_adisi(self):

        if self.isISI:
            return self.get_adtm()
        else:
            if self.nev > 0:
                tm = np.concatenate([np.array([self.tm[0]]), np.diff(self.tm)])
                return np.array([self.ad, tm])
            else:
                return np.zeros([2,0])

    def get_isiad(self):

        if self.isISI:
            return self.get_tmad()
        else:
            if self.nev > 0:
                tm = np.concatenate([np.array([self.tm[0]]), np.diff(self.tm)])
                return np.array([tm, self.ad])
            else:
                return np.zeros([2,0])

    def set_isi(self):
        if self.isISI:
            pass
        else:
            evs = self.get_adisi()
            self.set_data(evs[0],evs[1])
            self.isISI = True
            

    def set_abs_tm(self):
        '''
        Transform ISI timestamps into absolute time 
        '''
        if self.isISI:
            self.ad, self.set_tm(np.cumsum(self.get_tm()))
            self.isISI = False
        else:
            pass

    def filter_by_mapping(self, mapping):
        """
        Map the events, given a mapping dictionary like:
        map[src]=[target1,target2,...,targetn],
        """
        #The following is optimized for performance
        wasISI = False
        if self.isISI:
            wasISI = True
            self.set_abs_tm()
        evs = self.get_adtmev()
        ff = lambda x: x[0] in mapping
        filt = list(filter(ff, evs)) #keep only addresses that are mapped
        if len(filt) > 0:
            evs_filt = events(filt, self.atype).get_adtmev()
            #Get mapped addresses
            #list(chain(* concatenates lists of lists for low cost (i.e. O(n))
            m_ad = np.array(list(itertools.chain(*list(map(mapping.get, evs_filt[:, 0])))), self.dtype['ad'])
            #Get the respective timestamps
            m_tm = np.array(list(itertools.chain(*[len(mapping.get(x[0])) * [x[1]] for x in evs_filt])), self.dtype['tm'])

            self.set_data(m_ad,m_tm)
        else:
            self.empty()
        if wasISI:
            self.set_isi()

    def empty(self):
        self.__data = np.zeros([0], self.dtype)

    def iter_by_timeslice(self, tm):
        import bisect
        #ISI -> Cumulative
        if self.isISI:
            sum_tm = np.cumsum(self.get_tm())
        else:
            sum_tm = self.get_tm()
        #No in-place change
        id_start = 0
        t = 0

        #better to recycle an object rather than creating new (slow)
        evs = events(isISI=self.isISI)

        while id_start < len(sum_tm):
            t += tm
            id_stop = bisect.bisect_right(sum_tm, t, lo=id_start)
            evs.__data = self.__data[id_start:id_stop]
            id_start = id_stop
            rest = tm - evs.get_tdur()
            print((tm, evs.get_tdur()))

            if not evs.get_nev() > 0:
                rest = 0

            t -= rest
            #ISI or not is determined
            yield evs, tm - rest

    def __iter__(self):
        for ev in self.get_tmadev():
            yield ev[0],ev[1]

    def sort(self):
        '''
        Sort events by addresses and timestamps (in this order).
        '''
        self.set_abs_tm()
        self.__data.sort(order=['ad', 'tm'])

    def sort_tm(self):
        '''
        Sort events by timestamps and addresses (in this order).
        '''
        self.set_abs_tm()
        self.__data.sort(order=['tm', 'ad'])

    def demultiplex(self):
        '''
        Generates a dictionary with addesses as keys and a list of timestamps as values.
        Used internally for generating SpikeLists
        '''
        evs = events(ev=self)
        evs.sort()
        ads = np.unique(evs.ad)

        d = dict()
        k_start = 0
        k_stop = np.searchsorted(self.ad, ads, side='right')

        for i, a in enumerate(ads):
            d[a] = self.tm[k_start:k_stop[i]]
            k_start = k_stop[i]

        return d


class channelEvents(dict):
    '''
    inputs:
    *channel_events*: dictionary or channelEvents.
    *atype*: Address type 'physical' or 'logical'
    '''
    def __init__(self, channel_events=None, atype='physical'):
        assert isinstance(atype, str)
        self.__atype = atype.lower()[0]

        if isinstance(channel_events, dict):
            for k, v in list(channel_events.items()):
                self.add_ch(k, events(v, self.atype))
        elif channel_events is not None:
            raise TypeError("channel_events must be a dictionary, None or a channelEvents. Alternatively, use an events object and extract from channelAddressing")

    @property
    def atype(self):
        return self.__atype

    def copy(self):
        return channelEvents(self, atype=self.atype)

    def __add__(self, other):
        for i in list(other.keys()):
            if i in self:
                self[i] + other[i]
            else:
                self[i] = other[i]

    def __getattr__(self, attrName):
        """
        Check if the attribute exists in self, otherwise look for attributes of the encapsulated events and build a function by adding a channel argument to it.
        """
        if attrName in self.__dict__:
            return self.__dict__.__getitem__(attrName)
        else:
            def dnfun(ch, *args, **kwargs):
                return getattr(self[ch], attrName)(*args, **kwargs)
            return dnfun

    def add_ch(self, channel, ev):
        '''
        Add an events object to a channel
        Remark: Creates a reference to the provided events! In place changes also affect the events!
        '''
        assert channel not in self
        if not isinstance(ev, events):
            self[channel] = events(ev, atype=self.atype)
        else:
            self[channel] = ev

    def __len__(self):
        return self.get_nev()

    def add_adtmch(self, channel, ad, tm):
        if channel not in self:
            self[channel] = events(atype=self.atype)

        self.add_adtm(channel, ad, tm)

    def flatten(self):
        ev = events(atype=self.atype)
        for ch in self:
            ev.add_adtmev(self.get_adtmev(ch))
        return ev

    def filter_channel(self, channel_list=None):
        """
        Removes (in-place) all channels with are not in channel_list.
        If channel_list is omitted, the ch_events is left unchanged
        """
        if channel_list == None:
            return None
        all_ch = list(self.keys())
        for ch in all_ch:
            if ch not in channel_list:
                self.pop(ch)

    def get_all_tm(self):
        return self.flatten().get_tm()

    def get_last_tm(self):
        t = 0
        for evs in list(self.values()):
            t=max(t,evs.get_tm()[-1])
        return t

    def get_first_tm(self):
        t = np.inf
        for evs in list(self.values()):
            t=min(t,evs.get_tm()[0])
        return t

    def get_all_ad(self):
        return self.flatten().get_ad()

    def get_all_adtmev(self):
        return self.flatten().get_adtmev()

    def get_all_tmadev(self):
        return self.flatten().get_tmadev()

    def get_nev(self):
        n = 0
        for v in list(self.values()):
            n+=len(v)
        return n

    def iter_by_timeslice(self, tm):
        return self.flatten().iter_by_timeslice(tm)

    def filter_all_by_mapping(self, mapping):
        """
        Modifies, in-place, the encapsulated events acccording to the one-to-many mapping (key is int/float, value is iterable)
        """
        for ch in self:
            self[ch].filter_by_mapping(mapping)

    def filter_all_by_channel_mapping(self, mapping):
        """
        Modifies, in-place, the encapsulated events acccording to the one-to-many mapping (key is int/float, value is iterable)
        In this function, a different mapping is used for each channel (useful for address that do not contain channel information, for example)
        """
        for ch in self:
            self[ch].filter_by_mapping(mapping[ch])


def setDefaultMonChannelAddress(cs):
    """ Sets the default Monitoring Channel Addressing scheme (i.e. AER Output) in pyST globals, and used by pyAex. The argument is then returned when getDefaultMonChannelAddress is called

    cs: The Channel Addressing Scheme, instance of the channelAddressing class

    See also:
    `getDefaultMonChannelAddress <#pyST.STas.getDefaultMonChannelAddress>`_
    """

    pyST_globals.DefaultMonChannelAddress = cs
    return None


def getDefaultMonChannelAddress():
    """
    Returns the default Monitoring Channel Addressing Scheme (i.e. AER Output). No arguments

    See also:
    `setDefaultMonChannelAddress <#pyST.STas.setDefaultMonChannelAddress>`_
    """

    try:
        pyST_globals.DefaultMonChannelAddress
    except NameError:
        print('Error, run setDefaultMonChannelAddress first')
    return pyST_globals.DefaultMonChannelAddress


def setDefaultSeqChannelAddress(cs):
    """
    Sets the default Sequencer Channel Addressing scheme (i.e. AER Input) in pyST globals, and used by pyAex. The argument is then returned when getDefaultMonChannelAddress is called

    cs: The Channel Addressing Scheme, instance of the channelAddressing class

    See also:
    `pyST.getDefaultSeqChannelAddress <#pyST.STas.getDefaultSeqChannelAddress>`_
    """

    pyST_globals.DefaultSeqChannelAddress = cs
    return None


def getDefaultSeqChannelAddress():
    """
    Returns the default Monitoring Channel Addressing Scheme (i.e. AER Input). No arguments

    See also:
    `setDefaultSeqChannelAddress <#pyST.STas.setDefaultSeqChannelAddress>`_
    """

    try:
        pyST_globals.DefaultSeqChannelAddress
    except NameError:
        print('Error, run setDefaultSeqChannelAddress first')
    return pyST_globals.DefaultSeqChannelAddress


# Field class used for addrSpecIgnoreSynapseNeuron and STIgnoreNeuronSynapse
class channelAddressing:
    """
    This class contains the information defining the Channel Addressing Scheme for a multichip setup, i.e. definition of which chips are found on which channel

    channelAddressing(stasList,nChannelBits)

    *nChannelBits*: Positions of channel bits in the hardware addresses. Default nChannelBits=[15,16]

    *stasList*: list of address specification objects, of the same length as len(channelBits)**2

    See also:
    `addrSpec <#pyST.STas.addrSpec>`_

    """

    def __init__(self, stasList, channelBits=[15, 16]):
        """
        Constructor of the channelAddressing object

        See also
            SpikeTrain
        """
        self.nBitsTotal = np.array([0] * len(stasList), 'uint16')
        self.nChannelBits = len(channelBits)
        self.nChannels = 2 ** len(channelBits)
        self.channels = np.arange(self.nChannels, dtype='uint32')
        self.channelValues = 2**channelBits[0]*self.channels
        self.channelMask = np.sum(2**self.channels)
        self.channelShift = channelBits[0]
        self.addressMask = 2**channelBits[0]-1
        self.channelWidth = len(channelBits)

        if len(stasList) <= self.nChannels:
            for i in range(len(stasList)):
                if stasList[i] != None:
                    self.nBitsTotal[i] = stasList[i].nBitsTotal
                else:
                    self.nBitsTotal[i] = 0
        else:
            raise RuntimeError("len(stasList)>2**nBitChannel !")

        self.nBitsTotal = min(channelBits) - 1
        self.stasList = stasList

    def __getitem__(self, channel):
        return self.stasList[channel]

    def __len__(self):
        """
        Returns the length of the channel addressing scheme
        """

        return len(self.stasList)

    def __repr__(self):
        repr_str = "\n"
        for i in range(len(self)):
            repr_str += "Channel " + str(i) + ": " + self[i].__repr__() + "\n"
        return repr_str

    def __iter__(self):
        for addr_spec in self.stasList:
            yield addr_spec

    def reprAddressSpecification(self):
        '''
        String representation of the channel address specification
        '''
        base_label  = '|--Channel--|----Chip-----|\n'
        base_string = '|--{0}bits----|----{1}bits---|\n\n'.format(self.nChannelBits,self.nBitsTotal)
        chann_strings = []
        for n, addr_spec in enumerate(self):
            chann_strings.append('Channel {0}:\n'.format(n))
            chann_strings.append(addr_spec.repr_addr_spec(self.nBitsTotal))
            chann_strings.append('\n')
        repr_addr_spec = base_label+base_string+''.join(chann_strings)
        return repr_addr_spec


    def getValue(self, channel):
        """
        returns the channel mask

        channel: int between 0 and len(channelBits)**2
        """

        return np.uint32(self.channelValues[channel])

    def extract(self, ev):
        """
        Extracts the channel information from events. Returns a channelEvents object. Use addrPhysicalExtract for decoding the adresses instead.

        *eventsArray*: numpy array of hardware events (uint32)

        See also:
            addrPhysicalExtract
        """
        ch_events = channelEvents(atype='Physical')
        channel = (ev.get_ad() >> self.channelShift)&self.channelMask     # Get channel information.

        for channelIdx in range(self.nChannels):
            t = pylab.find(channelIdx == channel)
            # Much faster than boolean list or filter
            if len(t) > 0:
                ad = (ev.get_ad()[t]) 
                ch_events.add_adtmch(channelIdx, ad, ev.get_tm()[t])
        return ch_events

    def extract_channels(self, addr):
        """
        Extracts the channel information from an array of hardware events. Returns a channels. 

        *eventsArray*: numpy array of hardware events (uint32)

        See also:
            addrPhysicalExtract
        """
        return (addr >> self.channelShift)&self.channelMask


    def isChannelAddrList(self, addr):
        """
        Mostly internal function


        Check whether addr is a valid channelAddr list/dict

        *addr*: Either list of dimension channelAddressing.nChannels, containing arrays of addresses. The list index is the channel number, or dict whose keys are the channel numbers.

        See also:

            Stas.isValidAddress
        """
        if isinstance(addr, list):
            assert len(addr) == self.nChannels
        elif isinstance(addr, dict):
            possible_channels = list(range(self.nChannels))
            for i in list(addr.keys()):
                assert i in possible_channels, "The keys of the addr dict mst be valid channel numbers, i.e. between %d and %d" % (0, self.nChannels)

            addr_list = [None for i in range(self.nChannels)]
            for k, a in list(addr.items()):
                addr_list[k] = a
            addr = addr_list
        else:
            raise RuntimeError("addr must be a list or a dict")
        return addr

    def addrLogicalConstruct(self, addr):
        """
        Constructs Logical addresses, *i.e.* addresses in float format useful for plotting because they keep the neuron - synapse order.

        Returns a float numpy array.

        **NOTE:** Logical addresses do not contain channel information

        *addr*: address (in a form isChannelAddrList() can understand), such as ``{0:[range(15),2]}``

        See also:

            addrLogicalConstruct, addrLogicalExtract
        """
        addr = self.isChannelAddrList(addr)
        mainAddr = [None for i in range(self.nChannels)]

        #Construct channel by channel if not empty
        for channelIdx in range(len(addr)):
            if addr[channelIdx] is not None:
                mainAddr[channelIdx] = np.array(self[channelIdx]
                    .addrLogicalConstruct(addr[channelIdx]), 'float')

        return mainAddr

    def addrLogicalExtract(self, addr):
        """
        Extracts Logical Address with channel information

        *addr*: address (in a form isChannelAddrList() can understand), such as ``{0:[1.,2.,3.]}``
        """
        addr = self.isChannelAddrList(addr)
        mainAddr = [None for i in range(self.nChannels)]

        #Construct channel by channel if not empty
        for channelIdx in range(self.nChannels):
            if addr[channelIdx] is not None:
                addr[channelIdx] = np.array(addr[channelIdx], 'float')
                mainAddr[channelIdx] = self[
                    channelIdx].addrLogicalExtract(addr[channelIdx])

        return mainAddr

    def addrPhysicalConstruct(self, addr):
        """
        Constructs Physical addresses to human readable addresses

        *addr*: dictionary of human readable numbers ( the physical addresses ), with channel numbers as keys

        """
        addr = self.isChannelAddrList(addr)
        mainAddr = np.zeros([0], dtype='uint32')
        for channelIdx in range(self.nChannels):
            if addr[channelIdx] is not None:
                mainAddr = np.concatenate((
                        mainAddr,
                        self[channelIdx].addrPhysicalConstruct(
                            addr[channelIdx])
                        + self.getValue(channelIdx
                        )))
        #Add Channel Part
        return mainAddr

    def addrPhysicalExtract(self, addr):
        """
        Extracts Physical addresses to human readable addresses

        *addr*: numpy array of uint32 numbers ( the physical addresses )

        Output is channelAddrList, channel where channelEventsList is a list of arrays which contain the addresses and channel is an array of channels.
        """
        if not isinstance(addr, np.ndarray):
            addr = np.array(addr, np.uint32)
        if not addr.dtype == np.uint32:
            addr = addr.astype(np.uint32)

        channels_in_addr = self.extract_channels(addr)
        channelEventsList = [None for i in range(self.nChannels)]
        for channelIdx in np.unique(channels_in_addr):
            t = (channels_in_addr == channelIdx)
            channelEventsList[channelIdx] = self[channelIdx].addrPhysicalExtract( addr[t] & self.addressMask)
        return channelEventsList

    def addrPhysicalLogical(self, addr):
        """
        Extracts Physical addresses to logical addresses (directly)

        *addr*: numpy array of uint32 numbers ( the physical addresses )

        Output is channelAddrList, channel where channelEventsList is a list of arrays which contain the addresses and channel is an array of channels.
        """
        if not isinstance(addr, np.ndarray):
            addr = np.array(addr, np.uint32)
        if not addr.dtype == np.uint32:
            addr = addr.astype(np.uint32)

        channels= self.extract_channels(addr)
        channelEventsList = [None for i in range(self.nChannels)]
        for channelIdx in np.unique(channels_in_addr):
            t = (channels_in_addr == channelIdx)
            channelEventsList[channelIdx] =\
            self[channelIdx].addrPhysicalLogical( addr[t] & self.addressMask)
        return channelEventsList

    def importAER(self, input=None, sep='\t', dt=1e-6, format='a', isi=False, *args, **kwargs):
        """
        Function for extracting, translating events from a numpy array. Output is a channelEvents object.
        
        Inputs:
        *input*: if a string, will be treated as a filename and passed to np.loadtxt, if a numpy array, will be considered as events of dimension 2 x number of events. By default, the addresses are on [:,0]
        *format*: either 't' or 'a' respectively meaning timestamps and addresses on the first column. ('a' by default)
        *kwargs*: keyword arguments passed to np.loadtxt
        """

        if isinstance(input, str):
            try:
                if 'converters' in kwargs:
                    ae = np.loadtxt(input, **kwargs)
                else:
                    ae = np.loadtxt(
                        input, delimiter=sep, dtype='uint32', **kwargs)
            except IOError:
                ae = np.zeros([0, 2], 'uint32')

            if format == 'a':
                input = events(ae, 'p')
            elif format == 't':
                input = events(np.fliplr(ae), atype='Physical')
            else:
                raise RuntimeError("format must be a (addresses first) or t (timestamps first)")
        elif isinstance(input, channelEvents):
            pass
        else:
            input = events(input)

        if isi == True:
            input.set_tm(np.cumsum(input.get_tm()))

        ch_events = self.extract(input)
        ch_events_log = channelEvents(atype='Logical')

        for ch in list(ch_events.keys()):
            if len(ch_events.get_tm(ch)) > 0:
                ad = self[ch].addrPhysicalLogical(ch_events.get_ad(ch))
                ch_events_log.add_adtmch(ch, ad, ch_events.get_tm(ch))

        return ch_events_log

    def normalizeAER(self, ch_events):
        """
        Called before extract to throw away pre-stimulus data
        """
        if not isinstance(ch_events, channelEvents):
            raise TypeError("ch_events must be a channelEvents object")

        # Find the starting time. Empty spike trains are accepted with None and
        # shape[0]=0 arrays
        #Normalize
        if ch_events.get_nev() == 0:
            raise RuntimeError("ch_events does not contain any events!")

        tStartGlobal = ch_events.get_all_tm().min()
        for ch in ch_events:
            ch_events.set_tm(ch, ch_events.get_tm(ch) - tStartGlobal)
        return ch_events

    def generateST(self, ch_events, normalize=True):
        """
        Extracts events from eventsChannel, using the address specification specified for the given channels, returns a list of SpikeList
        """
        assert ch_events.atype == 'l', 'channelEvents must be of type logical'

        STStimOut = dict(list(zip(list(range(self.nChannels)), [SpikeList([], [])
             for i in range(self.nChannels)])))
        t_start = 0
        t_stop = 0
        if ch_events.get_nev() > 0:
            if normalize:
                ch_events = self.normalizeAER(ch_events)
                t_start = 0
            else:
                t_start = ch_events.get_all_tm().min() * 1e-3
            t_stop = ch_events.get_all_tm().max() * 1e-3

        if t_start == t_stop:
            t_stop += 1

        for ch in ch_events:
            stas = self[ch]
            #To ms
            ch_events.set_tm(ch, ch_events.get_tm(ch) / 1000)
            #eventsLogicalChannel[:,0]=eventsChannel[channelIdx][:,0]
            STStimOut[ch] = SpikeList(
                    ch_events.get_adtmev(ch),
                    ch_events.get_ad(ch),
                    t_start=t_start,
                    t_stop=t_stop
                    )

        return STStimOut

    def rawoutput_from_chevents(self, ch_events, func=None, normalize=True, filter_duplicates=False):
        """
        this function acts like generateST, but constructs a RawOutput object which delays the decoding until it is necessary.
        *ch_events* is a channelEvents object of type 'p' (Physical)

        Inputs:
            *func* - a dictionary of functions with channels as keys to decode the addresses. If omitted, all the channels are considered
            Outputs a RawOutput object
        """
        t_start = 0
        t_stop = 0

        ch_events = ch_events.copy()

        if func == None:
            # If no decoding functions are defined then use all of those
            # available in the channel addressing
            func_data_list = [None]*self.nChannels
            for i in range(self.nChannels):
                if self[i] != None:
                    func_data_list[i] = self[i].addrPhysicalLogical
            func_data = dict(list(zip(list(range(self.nChannels)), func_data_list)))
        else:
            func_data = func

        assert ch_events.atype == 'p', 'channel events must be of type physical (p)'

        if ch_events.get_nev() > 0:
            if normalize:
                ch_events = self.normalizeAER(ch_events)
                t_start = 0
            else:
                t_start = ch_events.get_first_tm() * 1e-3
            t_stop = ch_events.get_last_tm() * 1e-3

        if t_start == t_stop:
            t_stop += 1

        raw_data = {}

        for ch in ch_events:        
            ch_events.set_tm(ch, ch_events.get_tm(ch))
            raw_data[ch] = ch_events[ch]
        raw_out = RawOutput(
                raw_data,
                func_data,
                t_start=t_start,
                t_stop=t_stop,
                filter_duplicates=filter_duplicates
                )

        return raw_out

    def exportAER(self, spikeLists, filename=None, format='a', isi=True, sep='\t', addr_format='%u', time_format='%u', *args, **kwargs):
        '''
        spikeLists can be of the follwing type:
        - Monitors
        - SpikeList
        - list of SpikeLists of dimension nChannels
        - dictionary with channels as keys and SpikeLists as values
        - SpikeList is given, it will be interpreted as {0: spikeLists}.
        format specifies whether timestamps (format='t') or addresses (format='a') should be on the first column.
        *addr_format* and *time_format* format to be used by np.savetxt
        '''

        out = []
        assert format in ['t', 'a'], 'Format must be "a" or "t"'

        if hasattr(spikeLists, 'to_chstlist'):
            #Assuming it is of type Monitors
            spikeLists = spikeLists.to_chstlist()
        elif isinstance(spikeLists, list):
            assert len(spikeLists) == self.nChannels, "spikeLists must have dimension %d" % self.nChannels
            for i in range(len(spikeLists)):
                if not isinstance(spikeLists[i], SpikeList):
                    raise TypeError(
                        "Elements of spikeLists must be SpikeList objects!")

            spikeLists = dict(list(zip(list(range(len(spikeLists))), spikeLists)))
        elif isinstance(spikeLists, SpikeList):
            spikeLists = {0: spikeLists}
        elif isinstance(spikeLists, dict):
            for i in list(spikeLists.keys()):
                if not isinstance(spikeLists[i], SpikeList):
                    raise TypeError(
                        "Values of spikeLists must be SpikeList objects!")
        else:
            raise RuntimeError(
                "spikeLists must be either a: SpikeList, list or dict object")

        ev = events(atype='Physical')

        #Translate logical addresses to physical using a mapping
        tic = time.time()
        for ch in spikeLists:
            if isinstance(spikeLists[ch], SpikeList):
                slrd = spikeLists[ch].raw_data()
                if len(slrd) > 0:
                    #addr_present=spikeLists[ch].id_list()
                    # mapping=dict(zip(addr_present,self.addrPhysicalConstruct(
                    # {ch:self[ch].addrLogicalExtract(addr_present)})))
                    # Not using convert because it is very slow ( iterates over
                    # all events )
                    tmp_mapped_SL = np.fliplr(slrd)
                    mapped_SL = np.zeros_like(tmp_mapped_SL, dtype='uint32')
                    mapped_SL[:, 1] = tmp_mapped_SL[:, 1] * 1000  # ms
                    mapped_SL[:, 0] = self[ch].addrLogicalPhysical(tmp_mapped_SL[:, 0]) + self.getValue(ch)
                    ev.add_adtmev(mapped_SL)
                    # ev.add_adtmev(mapSpikeListAddresses(spikeLists[ch],mappin
                    # g).convert(format='[id,time*1000]'))
                else:
                    print("Warning: Empty SpikeList encountered")
        tictoc = time.time() - tic
        if PERFORMANCE_DEBUG:
            print(("Address encoding took {0} seconds".format(tictoc)))

        #Multiplex
        tic = time.time()
        sortedIdx = np.argsort(ev.get_tm())
        tictoc = time.time() - tic
        if PERFORMANCE_DEBUG:
            print(("Multiplexing took {0} seconds".format(tictoc)))

        tic = time.time()
        #Create new sorted events object
        if len(sortedIdx) > 0:
            ev = events(ev.get_adtmev()[sortedIdx, :], atype='p')
            #exportAER
            if isi:
                ev.set_isi()

        else:
            ev = events(atype='p')

        #Choose desired output: no filename given, return events
        if filename is None:
            return ev
        #Otherwise write to file
        else:
            #exportAER_file
            if format is 'a':
                np.savetxt(filename, ev.get_adtmev(),
                    fmt=addr_format + sep + time_format)
            elif format is 't':
                np.savetxt(filename, ev.get_tmadev(),
                    fmt=time_format + sep + addr_format)
            return ev

    def buildAllHashTables(self, channels=None):
        if channels is None:
            channels = self.channels
        for i in channels:
            addrBuildHashTable(self[i])

STChannelAddressing = channelAddressing


def addrLogicalConstruct(stas, addr):
    """
    Constructs Logical addresses, *i.e.* addresses in float format useful for plotting because they keep the neuron - synapse order.

    **NOTE:** Logical addresses do not contain channel information

    *addr*: address (in a form `isValidAddress <pyst.isValidAddress>`_ can understand)
    """

    #Check and Parse Address
    nEntries, addr = isValidAddress(stas, addr)
    #Initialize logical address vector (double)
    addrLogical = np.zeros([nEntries], 'float')

    #Integer component
    IntCmp = np.zeros([nEntries])
    #Fractional component
    FracCmp = np.zeros([nEntries])
    IntBitsUsed = np.zeros([nEntries])
    FracBitsUsed = np.zeros([nEntries])

    for hrf_index, hrf in enumerate(stas.iter_hrfs()):  # hrf_index
        #Sanity check
        assert np.all((addr[hrf_index, :] & (2 ** hrf['width'] - 1))
             == addr[hrf_index, :]), "Cropped significant bits"

        if hrf['type'] is 1:
            IntCmp = IntCmp + addr[hrf_index, :] * 2 ** IntBitsUsed
            IntBitsUsed = IntBitsUsed + hrf['width']

        elif hrf['type'] is -1:
            # Considered using length instead: creates rational numbers such as
            # 0.33333333 which does not look nice
            FracBitsUsed = FracBitsUsed + hrf['width']
            FracCmp = FracCmp + addr[hrf_index, :] * 2 ** (-FracBitsUsed)

    addrLogical = np.array(IntCmp + FracCmp, 'float')
    return addrLogical


def addrLogicalExtract(stas, addrLogical):
    """
    Description here
    """
    #TODO: Check and Parse Address
    addrLogical = np.array(addrLogical, 'float')
    nEntries = addrLogical.shape[0]
    FracBits = stas.nbits[-1]
    IntBits = 0

    #assert nEntries>0
    #Initialize logical address vector (double)
    addr = np.zeros([stas.nDims, nEntries], 'uint32')

    addr_frac, addr_int = np.modf(addrLogical)
    addr_int = addr_int.astype('int')
    addr_frac = (addr_frac * 2 ** stas.nbits[-1]).astype('int')

    for hrf_index, hrf in enumerate(stas.iter_hrfs()):  # hrf_index
        if hrf['type'] is 1:
            mask = (2 ** hrf['width'] - 1)
            addr[hrf_index, :] = (addr_int >> IntBits) & mask
            IntBits += hrf['width']
        elif hrf['type'] is -1:
            FracBits -= hrf['width']
            addr[hrf_index, :] = (addr_frac >> FracBits) & (
                2 ** hrf['width'] - 1)
            #decimal part

    return addr


def addrPhysicalConstruct(stas, addr):
    """
    Description here
    """

    #Check and Parse Address
    nEntries, addr = isValidAddress(stas, addr)

    #Initialize physical address vector (integer)
    addr = stas.addr_encoder.encode(addr)

    addrPhysical = [[]] * len(stas.field)
    for fieldIndex, field in enumerate(stas.iter_fields()):
        addrPhysical[fieldIndex] = field.construct(addr[fieldIndex])
    addrPhysical = np.sum(list(zip(*addrPhysical)), 1)

    #Build dictionary for quick reference
    stas.addrPhysicalExtract(addrPhysical)

    return np.array(addrPhysical, 'uint32')


def isValidPhysicalAddress(stas, addrPhys):
    """
    Checks if the given list, int, numpy array is a valid physical address and outputs a numpy array
    """

    #Parse type, output numpy array
    if      isinstance(addrPhys, int):
        addrPhys = np.array([addrPhys], 'uint32')
    elif isinstance(addrPhys, np.int):
        addrPhys = np.array([int(addrPhys)], 'uint32')
    elif isinstance(addrPhys, np.int32):
        addrPhys = np.array([int(addrPhys)], 'uint32')
    elif isinstance(addrPhys, np.int64):
        addrPhys = np.array([int(addrPhys)], 'uint32')
    elif isinstance(addrPhys, list):
        addrPhys = np.array(addrPhys, 'uint32')
    elif isinstance(addrPhys, np.ndarray):
        addrPhys = addrPhys.astype('uint32')
# assert addrPhys.dtype==np.dtype('uint32'), 'Error: physical addresses must be
# an integer array or an integer'
    else:
        addrPhys = np.array(addrPhys, 'uint32')
        raise TypeError("Type was %s" % type(addrPhys))

    if len(addrPhys) is 0:
        return np.array([], 'uint32')
    else:
        return addrPhys

def addrPhysicalExtract(stas, addrPhys):
    """
    addrPhysicalExtract takes a physical address as argument and returns a numpy array containing the human readable list of addresses. First dimension is the field type, second dimension is the address.
    Uses hash tables for accelerating look-up.
    """
    return addrPhysicalExtractDecode(stas, addrPhys)


def addrPhysicalLogical(stas, addrPhys):
    """
    Direct translation from Physical Addresses to Logical addresses using a hash table
    """
    addrPhys=isValidPhysicalAddress(stas, addrPhys)
    #failedIndex=np.where(stas.addrExtractLogicalFast[addrPhys]==-1)[0]
    failedIndex = np.setdiff1d(addrPhys, list(stas.addrExtractLogicalFast.keys()))
    #Don't even bother calling decode if everyone is in
    if len(failedIndex)>0:
       try:
           addrPhysicalLogicalDecode(stas, failedIndex) #Failed? decode the address
       except AssertionError as e:
           print('Error in addrPhysicalLogicalDecode.')
           print(('No. of addresses in the packet : {0}'.format(len(addrPhys))))
           raise e


# NOTE: If you get a ValueError, you probably have checkLevel=0 and forgot to
# build the hash tables using addrBuildHashTable
    fastAddr = np.array(
        list(map(stas.addrExtractLogicalFast.get, addrPhys)), 'float')

    # Loop should run over addresses only once if all addresse are in dict.
    # Small overhead
    return fastAddr.transpose()


def addrLogicalPhysical(stas, addrLogical, *args, **kwargs):
    """
    Direct translation from Logical Addresses to Physical addresses using a hash table
    """
    #Don't even bother calling decode if everyone is in
    try:
        # Failed? decode the address
        return addrLogicalPhysicalDecode(stas, addrLogical)  
    except AssertionError as e:
        print('Error in addrLogicalPhysicalDecode.')
        print(('No. of addresses in the packet : {0}'.
            format(len(addrLogical))))
        raise e

def _buildGrid(inlist):
    nD = len(inlist)
    min_list = [0] * nD
    max_list = [None] * nD
    for i in range(nD):
        max_list[i] = len(inlist[i])
    strmgrid = 'np.mgrid['
    for fieldIdx in range(nD):
        strmgrid = strmgrid + '{0}:{1},'.format(
            min_list[fieldIdx], max_list[fieldIdx])
    strmgrid = strmgrid + ']'
    allpo = eval(strmgrid)
    tot = np.prod(allpo[0].shape)
    grid = np.zeros([tot, len(allpo)], dtype=type(min_list[0]))
    grid_values = np.zeros_like(grid)
    for j in range(len(allpo)):
        grid[:, j] = allpo[j].flatten()
    for i in range(nD):
        grid_values[:, i] = np.array(inlist[i])[grid[:, i]]
    return grid_values


def addrBuildHashTable(stas):
    """
    addrBuildHashTable(stas) constructs all possible physical addresses of the given address specification and stores it internally in its hash table (stas.addrExtractFast). This function is typically used for real-time monitoring

    *stas:* addrSpec object.

    *Hash tables are not used anymore*

    This function does not scale well.
    """
    nDim = []
    for hrf in stas.iter_hrfs():
        nDim.append(hrf['range'])
    pass

    #Construct grid
    addrLens = [None] * len(stas)

    for i in range(len(stas)):
        addrLens[i] = stas[i]['range']

    grid = _buildGrid(addrLens)

    stas.allpos = stas.addrLogicalConstruct(grid.transpose())
    addr_phys = addrLogicalPhysicalDecode(stas, stas.allpos)
    addrPhysicalLogicalDecode(stas, addr_phys)


def addrPhysicalExtractDecode(stas, addrPhys):
    """
    addrPhysicalExtract takes a physical address as argument and returns a list containing the addresses in human readable form.
    This is called only the first time the physical address is decoded.

    *stas:* addrSpec object.

    *addrPhys*: an integer or an iterable containing integers representing physical addresses
    """

    #initalize address list
    if not hasattr(addrPhys, '__len__'):
        addrPhys = [addrPhys]

    addr = np.zeros([len(stas.field), len(addrPhys)], 'uint32')
    for fieldIndex, field in enumerate(stas.iter_fields()):
        addr[fieldIndex, :] = field.extract(addrPhys)
    addr = stas.addr_encoder.decode(addr)

    return addr


def addrPhysicalLogicalDecode(stas, addrPhys):
    '''
    Takes physical addresses and fills the physical to logical hash table
    '''
    addr = addrPhysicalExtractDecode(stas, addrPhys)
    addrLog = addrLogicalConstruct(stas, addr)
    #stas.addrExtractLogicalFast[addrPhys] = addrLog
    stas.addrExtractLogicalFast.update(list(zip(addrPhys, addrLog)))
    return addrLog


def addrLogicalPhysicalDecode(stas, addrLog):
    '''
    Takes logical addresses and fills the logical to physical hash table
    '''
    addr = addrLogicalExtract(stas, addrLog)
    addrPhys = addrPhysicalConstruct(stas, addr)
    stas.addrExtractPhysicalFast.update(list(zip(addrLog, addrPhys)))
    return addrPhys


def isValidAddress(stas, addrList):
    """
    This is an internal function which verifies that the given "human readable" address is consistent with the address specification. It also takes care of "filling in" the addresses: for example [range(15),2] is understood as [range(15),[2]*15].

    Raises AssertionError if addrList has incorrect data

    *stas*: an address specification object `addrSpec <pyST.addrSpec>`_
    *addrList*: a list of human readable addresses such as [range(15), 5]. It also accepts numpy arrays and transforms it accordingly
    """

        #    return len(addrList[0]),addrList
    if not isinstance(addrList, np.ndarray):
        assert getattr(addrList, '__iter__',
             False), "The addrList argument must be an iterable"
        # At the most minimum, check that the number of fields is consistent
        # with the length of the list
        assert len(addrList) == stas.nDims, "the number of fields is not consistent with the length of the list, are you using the correct address specification?"
        addrList = list(addrList)
    #make sure we have a 2 dimensional list and modify if necessary
        dimAddr = [0] * stas.nDims
        for hrf_index in stas.iter_hrf_index():
            if not hasattr(addrList[hrf_index], '__len__'):
                addrList[hrf_index] = [int(addrList[hrf_index])]
                #Make a list out of integer
            dimAddr[hrf_index] = len(addrList[hrf_index])

        #Remove all occurences of 1 in dimension list
        for i in range(dimAddr.count(1)):
            dimAddr.remove(1)

#        if not (len(np.unique(dimAddr)) in [0,1]):
# print("Dimensionalites of Address list is not consistent: Constructing grid")

        #Determine the number of entries
        if dimAddr != []:
            nEntries = np.prod(list(map(len, addrList)))
        else:
            nEntries = 1

        #Initialize address list
        addrListFilled = [[-1] * nEntries for i in range(stas.nDims)]

        addrListFilled = _buildGrid(addrList).astype('uint32').transpose(
            )  # not good make buildGrid use same type as in the min/max lists

    else:
        addrListFilled = addrList.astype('uint32')
        nEntries = addrListFilled.shape[1]

    #Range Check
    for hrf_index, hrf in enumerate(stas.iter_hrfs()):
        assert np.all(addrListFilled[hrf_index, :]
            >= 0), "Address must be a postive integer."
        diffs = np.setdiff1d(
            np.unique(addrListFilled[hrf_index, :]), hrf['range'])
        if len(diffs) != 0:
            try:
                err_ind = np.where(addrListFilled[hrf_index] == diffs)[0]
            except Exception as e:
                print((np.where(addrListFilled[hrf_index] == diffs), diffs))
                raise AssertionError(None, None)
            wrongaddr = addrListFilled[:, err_ind]
            print(("Address {3} is not in Range list ({0},{1}). Offending addresses on dimension {2}.".format(
                    np.max(hrf['range']),
                    np.max(hrf['range']),
                    hrf_index,
                    wrongaddr)))
            raise AssertionError(wrongaddr, err_ind)
    return nEntries, addrListFilled

class addressEncoder:
    def __init__(self):
        pass

    def encode(self, addr):
        return np.vstack(addr)

    def decode(self, addr):
        return np.vstack(addr)

    def encode_field(self, field, addr):
        return np.array(addr, 'uint32')

    def decode_field(self, field, addr):
        return np.array(addr, 'uint32')

class layoutFieldEncoder:  # private
    """
    Internal struct used for addrSpec, performing the pin layout permutation (the information contained in addrSr)
    """

    def __init__(self, shift, width, name = None):
        #Check if transformation is an identity function. If so, just mask and shift
        self.name=name
        mask = 2**width-1
        self.extract = lambda x: (x>>shift)&mask
        self.construct = lambda x: (x)<<shift

TYPE_TO_NAME_DICT = { -1 : 'synapse dimension', 0 : 'other dimension', 1 : 'neuron dimension'}

def repr_addr_spec(addr_spec, nBitsTotal):
    '''
    Prints the address specification in a human readable format
    *Input*: addr_spec, a AddrSpec object
    '''
    base_string = '| ' + ''.join(['{'+str(i)+'} ' for i in range(nBitsTotal)[::-1]]) +' |'
    tmp_str = [ 'I ']*nBitsTotal
    for k,a in list(addr_spec.addrSpec.items()):
        a_rev = a[::-1] #reverse
        for i in a_rev:
            tmp_str[i] = '{0}{1}'.format(k,str(i))
    p2a_map = base_string.format(*tmp_str)
    base_string = ['']
    tmp_str = '{0} {1} = {2} \n'
    for n, a in enumerate(addr_spec.addrPinConf):
        field = addr_spec[n]
        base_string.append( tmp_str.format( TYPE_TO_NAME_DICT[field['type']], field['id'], field['f']) )
    a2n_map = ''.join(base_string) + 'I = ignore \n'
    return p2a_map + ' \n' + a2n_map + ' \n'

class addrSpec:
    """
    Address specification class
    # NOTE : Documentation is outdated!!

    *addrConf*: should be a list of dicts. Each dict provides the address specificatoin information for one field, and should contain the following entries:

    - 'id' : whose value is equal to a single, unique, lowercase character.
    - 'range' : an iterable containing the possible values the address field can take
    - 'type': which is either 1 (Neuron) or -1 (Synapse) or -2 (Connection parameter).

    *addrStr*: should contain a list of space-separated pins with a number following an uppercase character defining the position on the chip. For example: "X0 X1 X2 X3 X4 X5 X6 Y4 Y3 Y2 Y1 Y0"

    """
    def __init__(self, addrStr='', addrConf=list(), addrPinConf='', name="NoName"):
        ##script for parsing the adddrStr
        self.id = name
        self.addrPinConf = addrPinConf
        self.addrStr = addrStr
        self.addrConf = addrConf
        #Useful to have the following functions as members:
        self.__class__.addrLogicalConstruct = addrLogicalConstruct
        self.__class__.addrLogicalExtract = addrLogicalExtract
        self.__class__.addrPhysicalConstruct = addrPhysicalConstruct
        self.__class__.addrPhysicalExtract = addrPhysicalExtract
        self.__class__.addrPhysicalLogical = addrPhysicalLogical
        self.__class__.addrLogicalPhysical = addrLogicalPhysical
        self.__class__.repr_addr_spec = repr_addr_spec
        #Update all parameters

        self.addrConf, self.addrDict, self.nBits, self.nBitsTotal = _process_addrConf(self.addrConf)
        self.addr_encoder = addressEncoder()
        self.nDims = len(self.addrConf)
        self.field, self.nFields = _stas_create_fields(self,self.nBits, self.addrConf)
        self.nbits = _stas_compute_nbits(self.addrConf)
        self.addrExtractLogicalFast = dict()
        self.addrExtractPhysicalFast = dict()

    def __len__(self):
        return self.nDims

    def __repr__(self):
        input_format = "[ "
        for v in self.iter_hrfs():
            input_format += str(v['id']) + ", "
        input_format += "]"

        return "AddrSpec: " + self.id + " " + input_format

    def iter_hrf_index(self):
        '''
        Iterates over the indexes of addrConf
        '''
        for k in range(self.nDims):
            yield k

    def iter_hrfs(self):
        '''
        Iterates over the list addrConf
        '''
        for v in self.addrConf:
            yield v

    def iter_fields(self):
        for field in self.field:
            yield field

    def iter_fields_by_pin(self, pin_order=None):
        if pin_order == None:
            pin_order = np.sort(list(self.nBits.keys()))
        for p in pin_order:
            for field in self.field:
                if field.pin == p:
                    yield p, field

    def __getitem__(self, field):
        return self.addrConf[field]

    def __getNHML__(self):
        '''
        Returns xml representation of this object
        '''
        doc = etree.Element('addressSpecification')
        # addrConf
        for conf in self.addrConf:
            # Dimensions
            dimdoc = etree.SubElement(doc, 'dim')
            # Id
            dimdoc.attrib['id'] = conf['id']
            # Type
            if conf['type'] == 1:
                dimdoc.attrib['type'] = 'soma'
            elif conf['type'] == -1:
                dimdoc.attrib['type'] = 'synapse'
            elif conf['type'] == -2:
                dimdoc.attrib['type'] = 'connection'
            else:
                # NOTE: In future if there are more types this is where they
                # are
                # to be added!
                dimdoc.attrib['type'] = 'unknown'
            # Range
            rangedoc = etree.SubElement(dimdoc, 'range')
            rangedoc.text = str(conf['range'])
            # Description
            descdoc = etree.SubElement(dimdoc, 'description')
            # Decoder
            decoderdoc = etree.SubElement(dimdoc, 'decoder')
        # Pins
        # Pinlayout
        return doc

def parse_addrSpec_NHML(doc):
    '''
    Parses an lxml element tree or a file name with xml content to
    initialize the object.
    '''
    if isinstance(doc, str):
        # parse the file
        doc = etree.parse(doc).getroot()
    else:
        # assuming doc is an lxml Element object.
        assert doc.tag == 'addressSpecification'

    addrConf = []  # NOTE: should probably be done in init function
    # addrConf
    for elm in doc:
        if elm.tag == 'dim':
            # Dimensions
            dim = {}
            # Id
            dim['id'] = elm.get('id')
            dim['description'] = elm.get('description')
            # Type
            if elm.get('type') == 'synapse':
                dim['type'] = -1
            elif elm.get('type') == 'soma':
                dim['type'] = 1
            elif elm.get('type') == 'connection':
                dim['type'] = -2
            
            else:
                # If there are any new types they should be added here!
                dim['type'] = None
            for chld in elm:
                if chld.tag == 'range':
                    # Range
                    if chld.text:
                        dim['range'] = eval(chld.text)
                    else:
                        dim['range'] = None
                else:
                    dim[chld.tag] = eval(chld.text)
            addrConf.append(dim)
    # Update the object
    return addrConf


#################################################################
#The following functions are used in AddrSpec
#
def _process_addrConf(addrConf):
    #Find number of required bits from range data
    for hrf in addrConf:
        bits_req = 1
        while max(hrf['range']) >> bits_req > 0:
            bits_req += 1
        if 'width' not in hrf:
            hrf['width'] = bits_req
        else:
            hrf['width'] = max(bits_req, hrf['width'])
        if 'shift' not in hrf:
            hrf['shift'] = 0
        if 'default_value' not in hrf:
            hrf['default_value'] = 0  # TODO: Add it to chipfile

    nBits = {}
    for i in addrConf:
            nBits[i['id']] = i['width']

    nBitsTotal = sum(nBits.values())
    #nBitsTotal = np.sum(nBits.values())

    addrDict = dict()
    #Find number of required bits from range data
    for i, hrf in enumerate(addrConf):
        addrDict[hrf['id']] = i

    return addrConf, addrDict, nBits, nBitsTotal


def _stas_create_fields(addrSpec,nBits,addrConf):
    """
    Creating fields for physical
    """
    nFields = len(nBits)
    field = [None for i in range(nFields)]
    pos = 0

    for nextfield in addrConf:
        currentfield = layoutFieldEncoder(
                        shift = nextfield['shift'],
                        width = nextfield['width'],
                        name = nextfield['id'],
                        )
        field[pos] = currentfield
        pos += 1

    return field, nFields


def _stas_compute_nbits(addrConf):
    nbits = {-1: 0, 1: 0, 0: 0, -2:0}
    for a in addrConf:
        nbits[a['type']] += a['width']
    return nbits

def extract_id_list(addr_conf):
    id_list = []
    for i in addr_conf:
        id_list.append(i['id'])
    return id_list


def load_stas_from_nhml(doc):
    '''
    load_stas_from_nhml(NHMLfile)
    This function returns addrSpec objects aerIn and aerOut by parsing the NHML
    file.
    '''
    if isinstance(doc, str):
        # parse the file
        doc = etree.parse(doc).getroot()
    else:
        # assuming doc is an lxml Element object
        assert doc.tag == 'chip'
    chipclass = doc.get('chipclass')
    for elm in doc:
        if elm.tag == 'addressSpecification':
            if elm.get('type') == 'aerIn':
                aerIn = addrSpec(name=chipclass + 'In', addrConf = parse_addrSpec_NHML(elm))
            elif elm.get('type') == 'aerOut':
                aerOut = addrSpec(name=chipclass + 'Out', addrConf = parse_addrSpec_NHML(elm))
            else:
                pass
    return aerIn, aerOut
    



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Emre Neftci
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------
# -*- coding: utf-8 -*-

#For jaer monitoring
import tarfile
import glob
import time
import os
import fnmatch
import warnings
try:
    import dill as pickle
except ImportError:
    warnings.warn('Cannot import dill, falling back to pickle. Several pyNCS objects cannot be pickled')
    import pickle 

import pylab
import numpy as np
from shutil import rmtree

### The globals class
class datacontainer:
    def __init__(self):
        self.directory = './'

global globaldata
globaldata = datacontainer()

REC_FN_SEQ = 'seq_address_specification'
REC_FN_MON = 'mon_address_specification'

def get_figsize(fig_width_pt, ratio='g'):
    """
    Method to generate figure size.
    """
    inches_per_pt = 1.0 / 72.0                # Convert pt to inch
    golden_mean = (np.sqrt(5) - 1.0) / 2.0    # Aesthetic ratio
    fig_width = fig_width_pt * inches_per_pt  # width in inches
    if ratio is 'g':
        fig_height = fig_width * golden_mean      # height in inches
    elif ratio is 's':
        fig_height = fig_width                  # square figure
    else:
        fig_height = 1. * fig_width / ratio
    fig_size = [fig_width, fig_height]      # exact figsize
    return fig_size


def loadPlotParameters(size=0.5, fontsize=18.0):
    """
    Load default matplotlib parameters for pretty plotting
    size: 0.5 -- two column page.
          0.33 -- three column page.
          0.25 -- two column double figure.
    fontsize: universal font size
    """
    if size <= 0.25:
        border = 0.22
    elif size <= 0.33:
        border = 0.20
    else:
        border = 0.15
    params0 = {'backend': 'pdf',
          'savefig.dpi': 300.,
          'axes.labelsize': fontsize,
          'figure.subplot.bottom': border,
          'figure.subplot.left': border,
          'text.fontsize': fontsize,
          'xtick.labelsize': fontsize,
          'ytick.labelsize': fontsize,
          #'legend.pad': 0.1,    # empty space around the legend box
          'legend.fontsize': fontsize,
          'lines.markersize': 5,
          'lines.linewidth': 1,
          'font.size': fontsize,
          'text.usetex': True,
          'figure.figsize': get_figsize(1000 * size, ratio=1.3)}  # size in inches
    pylab.rcParams.update(params0)
    
def load_compatibility(filename):
    """
    Same as experimentTools.load(), but works around recent module renaming problems
    Code from http://wiki.python.org/moin/UsingPickle/RenamingModules
    """
    import pickle as pickle #Does not work with dill
    renametable = {
        'pyST': 'pyNCS.pyST',
        'pyST.spikes': 'pyNCS.pyST.spikes',
        'pyST.STas': 'pyNCS.pyST.STas',
        'pyST.STsl': 'pyNCS.pyST.STsl',
        'pyST.stgen': 'pyNCS.pyST.stgen',
        }
    
    def mapname(name):
        if name in renametable:
            return renametable[name]
        return name
    
    def mapped_load_global(self):
        module = mapname(self.readline()[:-1])
        name = mapname(self.readline()[:-1])
        klass = self.find_class(module, name)
        self.append(klass)
        
    def loads(filename_):
        fh = file(filename_,'rb')
        unpickler = pickle.Unpickler(fh)
        unpickler.dispatch[pickle.GLOBAL] = mapped_load_global
        return unpickler.load()
    
    return loads(filename)

def load(filename=None, compatibility=False):
    """
    Unpickles file named 'filename' from the results directory. If no 'filename' is given, then 'globaldata.pickle' is loaded
    """
    if filename == None: 
        filename = globaldata.directory + 'globaldata.pickle'
    else:
        filename = globaldata.directory + filename        
    if compatibility:
        return load_compatibility(filename)
    else:
        return pickle.load(file(filename, 'rb'))    


def save_py_scripts():
    """
    Save all the python scripts from the current directory into the results directory
    """
    h = tarfile.open(globaldata.directory + 'exp_scripts.tar.bz2', 'w:bz2')
    all_py = glob.glob('*.py')
    for i in all_py:
        h.add(i)
    h.close()
    
def save_file(filename):
    """
    Save all the python scripts from the current directory into the results directory
    """
    import shutil
    shutil.copy(filename, globaldata.directory+filename)


def save(obj=None, filename=None):
    if obj == None and filename == None:
        f = file(globaldata.directory + 'globaldata.pickle', 'w')
        pickle.dump(globaldata, f)
        f.close()
        save_py_scripts()
    elif obj == None and filename != None:
        f = file(globaldata.directory + filename, 'w')
        pickle.dump(globaldata, f)
        f.close()
    else:
        f = file(globaldata.directory + filename, 'w')
        pickle.dump(obj, f)
        f.close()
    return None


def savetxt(obj, filename):
    np.savetxt(globaldata.directory + filename, obj)


def mksavedir(pre='Results/', exp_dir=None):
    """
    Creates a results directory in the subdirectory 'pre'. The directory name is given by ###__dd_mm_yy, where ### is the next unused 3 digit number
    """

    if pre[-1] != '/':
        pre + '/'

    if not os.path.exists(pre):
        os.makedirs(pre)
    prelist = np.sort(fnmatch.filter(os.listdir(pre), '[0-9][0-9][0-9]__*'))

    if exp_dir == None:
        if len(prelist) == 0:
            expDirN = "001"
        else:
            expDirN = "%03d" % (
                int((prelist[len(prelist) - 1].split("__"))[0]) + 1)

        direct = time.strftime(
            pre + "/" + expDirN + "__" + "%d-%m-%Y", time.localtime())
        assert not os.path.exists(direct)

    elif isinstance(exp_dir, str):
        direct = pre + exp_dir
        if os.path.exists(direct):
            print(("Warning: overwriting directory {0}".format(direct)))
            rmtree(direct)

    else:
        raise TypeError('exp_dir should be a string')

    os.mkdir(direct)

    globaldata.directory = direct + str('/')

    fh = file(
        globaldata.directory + time.strftime("%H:%M:%S", time.localtime()), 'w')
    fh.close()

    print(("Created experiment directory {0}".format(globaldata.directory)))
    return globaldata.directory


def savefig(filename, *args, **kwargs):
    """
    Like pylab.savefig but appends the Results directory
    """
    pylab.savefig(globaldata.directory + filename, *args, **kwargs)


def annotate(filename='', text=''):
    "Create a file in the Results directory, with contents text"
    f = file(globaldata.directory + filename, 'w')
    f.write(text)
    f.close()

def save_rec_files(nsetup):
    '''
    Saves files recorded by the communicator and prepends address specification
    '''
    fh_addr_spec_mon = file(globaldata.directory + REC_FN_SEQ, 'w')
    fh_addr_spec_seq = file(globaldata.directory + REC_FN_MON, 'w')
    seq_addr_spec_str = nsetup.seq.reprAddressSpecification()
    mon_addr_spec_str = nsetup.mon.reprAddressSpecification()
    fh_addr_spec_mon.write(mon_addr_spec_str)
    fh_addr_spec_seq.write(seq_addr_spec_str)
    fh_addr_spec_mon.close()
    fh_addr_spec_seq.close()
    exp_fns = nsetup.communicator.get_exp_rec()
    import shutil, os
    for f in exp_fns:
        shutil.copyfile(f, globaldata.directory+f.split('/')[-1])
        
def savefigs(filename='fig', extension = 'png', close=True, *args, **kwargs):
    """
    Saves all figures with filename *filename#* where # is the figure number.
    The order is: last opened last saved.
    Inputs:
    *filename*: figure name prefix
    *extension*: figure extension. savefig should resolve the format from the extension
    *close*: whether to close the figure after it is saved
    *args, **kwargs  are passed to savefig  
    """
    import matplotlib,pylab
    figures = [manager.canvas.figure
         for manager in matplotlib._pylab_helpers.Gcf.get_all_fig_managers()]
    print(('Saving {0} figures'.format(len(figures))))
    for i, f in enumerate(figures):
        savefig('fig'+str(i)+'.'+extension, *args, **kwargs)
        if close: pylab.close()



#!/bin/python
#-----------------------------------------------------------------------------
# File Name : testComAPI.py
# Purpose:
#
# Author: Emre Neftci
#
# Creation Date : 05-06-2013
# Last Modified : Wed 05 Jun 2013 06:47:18 PM PDT
#
# Copyright : (c) 
# Licence : GPLv2
#----------------------------------------------------------------------------- 

from pyNCS.ComAPI import *

class Communicator(BatchCommunicatorBase):
    def run(self, stimulus=None, duration=None, context_manager=None):        
        return stimulus




#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Emre Neftci
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

import os
import numpy as np
import pylab
import matplotlib
from . import pyST_globals
from . import STgen as stgen
from .spikes import SpikeList, SpikeTrain, load, merge, merge_spikelists

#Globals
addrIndex = 0
timeIndex = 1
STCreate = stgen.StGen()


def plot_raster(SL, id_list=None, t_start=None, t_stop=None, display=True, id_color=None, kwargs={}):
    ''' Same as spikelist.raster_plot() but with pretty plot options

        id_color is used to customize colors according to neuron address (for example inhibitory vs. excitatory)

        Example:
        >>> id_color = [{'ids':range(124),'color':'blue'},{'ids':range(124,128),'color':'red'}]
        >>> plot_raster(chipout[158],id_color=id_color)
    '''
    #SL._SpikeList__calc_startstop()
    kwargs_default = {'marker': '|', 'markersize': 2, 'color': 'black'}
    kwargs_default.update(kwargs)

    if id_color == None:
        SL.raster_plot(id_list, t_start, t_stop, display, kwargs)
    else:
        h = pylab.axes()
        for i in range(len(id_color)):
            kwargs['color'] = id_color[i]['color']
            SL.id_slice(id_color[i]['ids']).raster_plot(
                t_start=t_start, t_stop=t_stop, display=h, kwargs=kwargs)
        pylab.ylim([SL.id_list().min() - 1, SL.id_list().max() + 1])


STPlotRaster = plot_raster


def composite_plot_superimposed(st_pre, st_post, kwargs={}, colors=['blue', 'k'], id_list=None, t_start=None, t_stop=None, t_start_rate=None, t_stop_rate=None, pre_rate_mult=1.):
    """
    Make a nice Composite plot superimposing two spike trains, a raster plot combined with a vertical rate plot.
    *st_pre* and *st_post* are two SpikeLists. st_post will be plotted above st_pre
    *colors* is a list specifying the respective colors for the bars and spikes for st_pre and st_post
    *kwargs* are the keyword arguments for common plot arguments in the raster plots. Default is {'alpha':0.7,'marker':','} for st_pre and plot_raster defaults for st_post
    *pre_rate_mult* is a float used for plotting the rates of st_pre on a different scale
    other arguments are the same as in composite_plot
    """
    kwargs_pre = dict()
    kwargs_pre.update({'alpha': 0.8, 'marker': ',', 'color': colors[0]})
    kwargs_pre.update(kwargs)
    kwargs_post = dict()
    kwargs_post.update(kwargs)
    kwargs_post.update({'color': colors[1]})

    h1, h2 = composite_plot(
            st_pre,
            kwargs=kwargs_pre,
            kwargs_bar={'alpha': 0.8, 'color': colors[0]},
            id_list=id_list,
            t_start=t_start,
            t_stop=t_stop,
            t_start_rate=t_start_rate,
            t_stop_rate=t_stop_rate,
            )

    twin = False
    if pre_rate_mult != 1.:
        h2t = h2.twiny()
        twin = True
    else:
        h2t = h2

    super_plot = composite_plot(
            st_post,
            kwargs=kwargs_post,
            kwargs_bar={'color': colors[1]},
            display=[h1, h2t],
            id_list=id_list,
            t_start=t_start,
            t_stop=t_stop,
            t_start_rate=t_start_rate,
            t_stop_rate=t_stop_rate)

    if twin:
        pre_lims = h2.get_xlim()
        #h2.set_xlim(pre_lims[0],pre_lims[1]*pre_rate_mult)
        color = colors[0]
        for tl in h2.get_xticklabels():
            tl.set_color(color)
        h2.set_xlabel('Frequency (Hz)', color=color)

    return super_plot


def composite_plot(SL, id_list=None, t_start=None, t_stop=None, t_start_rate=None, t_stop_rate=None, display=True, kwargs={}, kwargs_bar={}):
    """
    Make a nice Composite plot, *i.e.* a raster plot combined with a vertical rate plot.

    The arguments are identical to STPlotRaster, except for display.
    *display* If True is given a new figure is created. If [axS,axR] is given, where axS and axR are pylab.axes objects, then the spike rater and the rate plot are plotted there.
    """
    SL._SpikeList__calc_startstop()

    if id_list == None:
        id_list = SL.id_list()
    if t_start is None:
            t_start = SL.t_start
    if t_stop is None:
            t_stop = SL.t_stop
    if t_start_rate is None:
            t_start_rate = t_start
    if t_stop_rate is None:
            t_stop_rate = t_stop

    if display == True:
        h = pylab.figure()
        axS = pylab.axes([0.12, 0.12, 0.57, 0.8])
        axR = pylab.axes([0.75, 0.12, 0.20, 0.8])
    elif isinstance(display, list):
        axS = display[0]
        axR = display[1]

    STPlotRaster(SL=SL, id_list=id_list, t_start=t_start, t_stop=t_stop,
         display=axS, kwargs=kwargs)

    min_addr = int(np.min(SL.id_list()))
    max_addr = int(np.max(SL.id_list()))
    axS.set_xlim([t_start, t_stop])
    axS.set_yticks([min_addr, (max_addr - min_addr) / 2, max_addr])
    axS.set_xticks(
        [int(t_start), 100 * (int(t_stop - t_start) / 200), int(t_stop)])

    rates = np.array(SL.mean_rates(t_start=t_start_rate, t_stop=t_stop_rate))
    barplot = axR.barh(id_list, rates, linewidth=0, **kwargs_bar)
         #remove space between bars, remove black lines
    pre_lims = axS.set_ylim()
    axR.set_ylim(pre_lims[0], pre_lims[1])

    axR.set_yticks([])
    axR.grid('on')
    #axS.xaxis.grid('on')

    axR.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(2))
    pylab.axes(axR)
    axR.set_xlabel('Frequency (Hz)')

    return axS, axR

STCompositePlot = composite_plot


def composite_plot_movie(SL, time_bin=10, t_start=None, t_stop=None, output="animation.mpg", bounds=(0, 5), fps=10, display=True, maxrate=None, ratebin=None, kwargs={}):
    pylab.ioff()
    from NeuroTools.plotting import get_display
    subplot = get_display(display)
    assert os.system('mencoder') != 32512, "mencoder not found!"
    if t_start is None:
        t_start = SL.t_start
    if t_stop is None:
        t_stop = SL.t_stop
    if maxrate is None:
        maxrate = 100
    if ratebin is None:
        ratebin = 100

    files = []
    im = pylab.figure(**kwargs)
    count = 0
    idx = 0
    axS, axR = STCompositePlot(SL, t_start=0, t_stop=time_bin,
         t_start_rate=0, t_stop_rate=ratebin, kwargs=kwargs)
    axR.hold(False)
    axS.hold(False)
    if t_start != SL.t_start or t_stop != SL.t_stop:
        spk = SL.time_slice(t_start, t_stop)
        raise RuntimeError('Not implemented')

    while (t_start < t_stop):
        STCompositePlot(SL, t_start=0, t_stop=t_start + time_bin, t_start_rate=t_start,
             t_stop_rate=t_start + ratebin, display=[axS, axR], kwargs=kwargs)
        axS.set_xlim([0, t_stop])
        axR.set_xlim([0, maxrate])
        fname = "_tmp_spikes_%05d.png" % count
        pylab.savefig(fname)
        files.append(fname)
        t_start += time_bin
        count += 1
        print(('Generated frame {0}'.format(count)))
    command = "mencoder 'mf://_tmp_*.png' -mf type=png:fps=%d -ovc lavc -lavcopts vcodec=wmv2 -oac copy -o %s" % (fps, output)
    os.system(command)
    for fname in files:
        os.remove(fname)


def mapSpikeListAddresses(SL, mapping=None):
    '''
    this function maps the addresses of a spike list into another using the given mapping. Useful for logical to physical translation and vice versa
    SL=original spike list
    mapping=dictionary containing address mapping, If mapping is None, then the addresses will be mapped onto a linear scale, i.e. range(len(SL.id_list()))
    '''
    assert isinstance(SL, SpikeList), "SL must be a NeuroTools SpikeList"
    mapped_SL = SpikeList([], [])
    addr_SL = SL.id_list()
    if mapping == None:
        mapping = dict(list(zip(np.sort(SL.id_list()), list(range(len(SL.id_list()))))))
    for k, v in list(mapping.items()):
        if k in addr_SL:
            try:
                mapped_SL[v] = SL[k]
            except KeyError:
                pass

    all_others = np.setdiff1d(SL.id_list(), list(mapping.keys()))
    for k in all_others:
        mapped_SL[k] = SL[k]

    return mapped_SL


def ksi(SL, t_stop=800, pow_freq=None, t_bin=10.):
    """
    Kuramoto Synchronization Index (KSI) (Kuaramoto 1984) for measuring phase coherency with respect to the strongest    freqeuncy of the population activity.

    Returns ksi in the range (0,1). 0 meaning no cohrency and 1 meaning entirely coherent

    Inputs:
        SL      - SpikeList of duration of at least t_stop
        t_stop  - the end of the SpikeTrain (in ms)
        t_bin   - time bin for computing population activity and spectrum
        pow_freq- if given, this frequency is taken as the strongest frequency

    """
    #Determine the strongest frequency
    N = len(SL)

    if pow_freq == None:
        psth = SL.time_slice(
            t_start=0, t_stop=t_stop).spike_histogram(t_bin).mean(axis=0)
        t = np.arange(0, t_stop * 1e-3, t_bin * 1e-3)
        sp = abs(np.fft.fft(psth).real)
        freq = abs(np.fft.fftfreq(t.shape[-1], d=t_bin * 1e-3))
        pow_freq = freq[1 + np.argmax(sp.real[1:])]

    l_vs = [[] for i in range(len(SL))]

    for i, k in enumerate(SL.id_list()):
        tm = SL[k].spike_times / 1000  # In seconds
        l_vs[i] = (tm + .5 / pow_freq) % (1. / pow_freq) - .5 / pow_freq

    vs = np.hstack(l_vs) * pow_freq * 2 * np.pi  # Flatten

    return abs(np.sum(np.exp(1j * vs)) / len(vs)), pow_freq



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: <authors name>
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------
#ConfAPI
#Biases and mapper
#Api for modules having pyAMDA-like functionality
#Api for modules having pyAEX-like functionality

from types import GeneratorType  # For checking generator type
from contextlib import contextmanager
from lxml import etree
import warnings

class Parameter:
    def __init__(self, param_name, params_dict):
        '''
        Parameter(parameters, configurator)
        params_dict: dictionary of parameters and values
        This object is designed to be used with the configurator to set parameters
        '''
        self.param_data = dict(parameters)
        self.SignalName = param_name

    def __str__(self):
        return str(self.param_data)





#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Sadique Sheik
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

from xml.dom import minidom as md
try:
    from urllib.request import urlopen
    from urllib.error import URLError, HTTPError
except:
    from urllib.request import urlopen, URLError, HTTPError
    
from .chip_v2 import NeuroChip
from .mapping import Mapping, PMapping
from .monitors import Monitors
from . import pyST
import warnings, os
from contextlib import contextmanager
from lxml import etree
from itertools import chain
from .api import ComAPI
from .api import ConfAPI
import numpy as np

_ROOT = os.path.abspath(os.path.dirname(__file__))
def get_data(path):
    return os.path.join(_ROOT, 'data', path)

CHIPFILESDIR ='NHML/' 
URL_SETUPTYPEDTD = get_data('setuptype.dtd')
URL_TIMEOUT = 0.3
# temp

#Useful function to get path to package_data


def dict_merge(x, y):
    '''
    Convenience function to merge two dictionaries
    returns the merged dictionary
    '''
    temp = x.copy()
    temp.update(y)
    return temp


def xml_parse_parameter(node):
    '''
    Convenience function to parse parameter leaves of XML
    '''
    d = dict()
    iterparams = node.iterfind('parameter')
    for i in iterparams:
        d[str(i.attrib['name'])] = i.text
    return d


def parse_and_validate(filename, dtd, validate = True):
    #handle internet urls appropriately
    if 'http://' in filename:
        doc = etree.parse(urlopen(filename))
    else:
        doc = etree.parse(filename)
    
    if validate:
        parser = etree.DTD(open(dtd))
        parser.validate(doc)
        if len(parser.error_log.filter_from_errors()) > 0:
            for e in parser.error_log.filter_from_errors():
                print(e)
            raise Exception('setupfile XML is not well formed, correct the errors given above first')
        elif len(parser.error_log) > 0:
            for e in parser.error_log.filter_from_errors():
                print(e)

    return doc


def get_addrspec(node, typ):
    if node == None:
        return []
    else:
        if typ in node.attrib:
            return list(eval(node.attrib[typ]))
        else:
            return []

class chAddrSpecs(object):
    '''
    Class that internally encapsulates the channelAddressing and chipslots.
    '''
    def __init__(self, mon=None, seq=None, chipslots={}, aerSlot={}):
        self.mon = mon
        self.seq = seq
        self.chipslots = chipslots
        self.aerSlot = aerSlot
        self.chip_aerin = {}
        self.chip_aerout = {}
        
    def get_chip_aerin(self, chipid):
            return self.chip_aerin[chipid]

    def get_chip_aerout(self, chipid):
            return self.chip_aerout[chipid]

class NeuroSetup(object):
    def __init__(self,
            setupfile,
            com_kwargs={},
            map_kwargs={},
            conf_kwargs={},
            prefix='',
            offline=False,
            validate = True):
        '''
        Inputs:
        *setuptype*: Defines the channels and the slots in the setup. See NCS documentation.
        *setupfile*: Indicates which chips occupy the slots in the setup. See NCS documentation.
        *com_kwargs*: keyword arguments for communicator module. This gets merged with arguments in setup.xml
        *map_kwargs*: keyword arguments for communicator module. This gets merged with arguments in setup.xml
        *conf_kwargs*: keyword arguments for communicator module. This gets merged with arguments in setup.xml
        *prefix*: path prefix to chipfiles (default: path to files distributed through pyNCS_data package_data)
        *offline*: if True, the setup will not communicate, map or configure.
        '''
        #Initialize and save API
        self.chaddrspecs = chAddrSpecs()
        self.chipslots = self.chaddrspecs.chipslots
        self.aerSlot = self.chaddrspecs.aerSlot
        self.get_chip_aerout = self.chaddrspecs.get_chip_aerout
        self.get_chip_aerin = self.chaddrspecs.get_chip_aerin

        self.com_kwargs = com_kwargs
        self.conf_kwargs = conf_kwargs
        self.map_kwargs = map_kwargs
        #Save setup file locations
        self.setupfile = setupfile
        self.prefix = prefix
        self.chips = {}
        self.slots = {}
        self.offline = offline
        self.validate = validate

        self.load_setup(self.setupfile, validate = validate, offline = offline)
        self.update()
        self.apply()
        self.reset()
        self.monitors = Monitors()
        self.sequencers = Monitors()

        if self.offline:
            warnings.warn('Running in Offline mode')

    @property
    def mon(self):
        return self.chaddrspecs.mon

    @mon.setter
    def mon(self, mon_ch_addr):
        self.chaddrspecs.mon = mon_ch_addr

    @property
    def seq(self):
        return self.chaddrspecs.seq

    @seq.setter
    def seq(self, seq_ch_addr):
        self.chaddrspecs.seq = seq_ch_addr

    def get_chipfilename(self, filename):
        if self.prefix == '':
            chipfile = get_data(str(CHIPFILESDIR+filename))
        else:
            chipfile = self.prefix + str(filename)
        return chipfile

    def load_setup(self, filename, validate = True, offline=False):
        '''
        Loads the setup
        Inputs:
        *filename*: setup file name
        *offline*: if True, the chips will not be configured ("pretend" mode).
        '''

        nsetup = parse_and_validate(filename, dtd=URL_SETUPTYPEDTD, validate = validate)
        for n in nsetup.iterfind('channelAddressing'):
            if str(n.attrib['type']) == 'monitor':
                self.monBits = np.array(eval(n.attrib['bits']), dtype='uint32')
                self.monChannels = np.array(self.monBits-self.monBits[0], dtype='uint32')
            elif str(n.attrib['type']) == 'sequencer':
                self.seqBits = np.array(eval(n.attrib['bits']), dtype='uint32')
                self.seqChannels = np.array(self.seqBits-self.seqBits[0], dtype='uint32')
            else:
                raise TypeError('channelAddressing type should be either monitor or sequencer')

        #parse defaultchip (should be unique)
        #self.defaultchipfile = self.get_chipfilename(str(nsetup.find('defaultchip').attrib['chipfile']))
        #
        # Load communicator, currently, only only communicator per setup is
        # supported
        for ncom in nsetup.iterfind('communicator'):
            if offline:
                self.com_api = ComAPI
                self.com_kwargs = {}
            else:
                com_kwargs = xml_parse_parameter(ncom)
                self.com_kwargs = dict_merge(self.com_kwargs, com_kwargs)
                try:
                    self.com_api = self._import_module(str(ncom.attrib['module']))
                except ImportError as e:
                    warnings.warn('Communicator module:' + e.message)
                    self.com_api = ComAPI
                    self.com_kwargs = {}
            self.communicator = self.com_api.Communicator(**self.com_kwargs)
            self.communicator.register_neurosetup(self)
        
        #Load configurators and chips
        for nchip in nsetup.iterfind('chip'):
            chipid = str(nchip.attrib['id'])
            slot = int(eval(nchip.attrib['slot']))
            chipfile = self.get_chipfilename(str(nchip.attrib['chipfile']))

            nconf = nchip.find('configurator')
            module = str(nconf.attrib['module'])
            conf_kwargs = xml_parse_parameter(nconf)
            try:
                conf_api = self._import_module(module)
            except ImportError as e:
                warnings.warn('Configurator module' + e.message)
                conf_api = ConfAPI
                conf_kwargs = {}
            chip = NeuroChip(chipfile, id=chipid, offline=offline,
                             conf_api=conf_api,
                             conf_kwargs=conf_kwargs)
            self.chips[chipid] = chip
            self.chipslots[chipid] = slot
            self.slots[slot] = chipid
            self.chaddrspecs.chip_aerin[chipid] = chip.aerIn
            self.chaddrspecs.chip_aerout[chipid] = chip.aerOut
            chip.configurator.register_neurosetup(self)

        #Load Mapper
        # load a virtual mapping table by defauld
        self.mapping = Mapping('Virtual Mapping') 

        for nmapper in nsetup.iterfind('mapper'):
            map_kwargs = xml_parse_parameter(nmapper)
            self.map_kwargs = dict_merge(self.map_kwargs, map_kwargs)
            try:
                self.map_api = self._import_module(str(nmapper.attrib['module']))
            except ImportError as e:
                warnings.warn('Mapper module:' + e.message)
                self.map_api = ConfAPI
                self.map_kwargs = {}
            self.mapper = self.map_api.Mappings(**self.map_kwargs)
            if 'version' in self.map_kwargs:
                if float(self.map_kwargs['version']) == 3.0:
                    self.mapping = PMapping('mapping')
                else:
                    self.mapping = Mapping('mapping')

        #If there is no mapper tag in setup, build an empty mapper
        if not hasattr(self, 'mapper'):
            from .api.ConfAPI import Mappings
            self.mapper = Mappings()
        self.mapper.register_neurosetup(self)

    def _import_module(self, module):
        try:
            #without fromlist, only the package is returned
            mod = __import__(module, fromlist=[module])
        except ImportError as e:
            #print(e)
            raise ImportError('{0} API failed to be imported'.format(module))
        return mod

    def update(self):
        '''
        updates the default monitor/sequencer (pyST) with the chips contained
        in this setup -- always call this function when finished with adding chips
        '''
        self.chipsIn = [None for i in range(2 ** len(self.seqBits))]
        self.chipsOut = [None for i in range(2 ** len(self.monBits))]
        seqList = [None for i in range(2 ** len(self.seqBits))]
        monList = [None for i in range(2 ** len(self.monBits))]

        for id, nslot in list(self.chipslots.items()):
            #Consider defining a function (for esthetic reasons)
            self.aerSlot[nslot] = dict()
            ######### Mon
            self.aerSlot[nslot]['monIn'] = [nslot]
            self.aerSlot[nslot]['monOut'] = [nslot]

            ######### Seq
            self.aerSlot[nslot]['seqIn'] = [nslot]
            self.aerSlot[nslot]['seqOut'] = [nslot]

#            for ii in self.aerSlot[nslot]['monIn']:
#                try:
#                    if self.chips[id].aerIn is not None:
#                        if  not self.chips[id].virtual:
#                            monList[ii] = self.chips[id].aerIn
#                            self.chipsIn[ii] = self.chips[id]
#                except KeyError as e:
#                    pass
#                #If mon.aerIn doesn't exist, then skip it
#                except AttributeError as e:
#                    pass

            for ii in self.aerSlot[nslot]['monOut']:
                try:
                    if self.chips[id].aerOut is not None:
                        if  not self.chips[id].virtual:
                            monList[ii] = self.chips[id].aerOut
                            self.chipsOut[ii] = self.chips[id]
                except KeyError as e:
                    raise e
                except AttributeError as e:
                    pass

            for ii in self.aerSlot[nslot]['seqIn']:
                try:
                    if self.chips[id].aerIn is not None:
                        if not self.chips[id].virtual:
                            seqList[ii] = self.chips[id].aerIn
                            self.chipsIn[ii] = self.chips[id]
                except KeyError as e:
                    raise e
                except AttributeError as e:
                    pass
                #If seq.aerIn doesn't exist, then skip it

#            for ii in self.aerSlot[nslot]['seqOut']:
#                try:
#                    if self.chips[id].aerOut is not None:
#                        if  not self.chips[id].virtual:
#                            seqList[ii] = self.chips[id].aerOut
#                            self.chipsOut[ii] = self.chips[id]
#                except KeyError as e:
#                    raise e
#                except AttributeError as e:
#                    pass
        self.seq = pyST.channelAddressing(channelBits=self.seqBits, stasList=seqList)
        self.mon = pyST.channelAddressing(channelBits=self.monBits, stasList=monList)

    def apply(self):
        ''' sets default monitor/sequencer to this setup '''
       #?? self.update()
        pyST.setDefaultMonChannelAddress(self.mon)
        pyST.setDefaultSeqChannelAddress(self.seq)

    def reset(self):
        for chip in list(self.chips.values()):
            chip.configurator.reset()

    def prepare(self):
        self.mapping.prepare()
        if not self.offline:
            if len(self.mapping.mapping) > 0:
                self.mapper.set_mappings(self.mapping.mapping)

    def __copy__(self):
        return self.__class__(self.setupfile,
                              prefix=self.prefix, offline=self.offline)

    def __deepcopy__(self, memo):
        return self.__copy__()

    def __getstate__(self):
        return {
                "prefix" : self.prefix,
                "offline" : self.offline,
                "validate" : self.validate,
               }

    def reload(self):
        '''
        Call this function if you recovered the object from a pickle and doesnt
        have any of the information or the information is corrupted.
        '''
        self.com_kwargs = {}
        self.conf_kwargs = {}
        self.map_kwargs = {}
        self.chips = {}
        self.chipslots = {}
        self.slots = {}
        self.load(self.setupfile, offline = self.offline, validate = self.validate)
        self.update()

    def _pre_process(self, stim):
        if stim == None:
            stim = self.sequencers
        evs_in = self.mon.exportAER(stim, isi=True)
        evs = self.mapper.filter_events(evs_in)
        return evs.get_adtmev()

    def _post_process(self, evs, filter_channels=None):
        evs_out = pyST.events(evs, 'p')
        mon_ch_addr = self.mon
        #extract per channel events -> ch_events
        ch_evs = mon_ch_addr.extract(evs_out)
        #Filter
        #-Channel Events filtering is easier to use than RawOutput.decode_dict
        if filter_channels != None:
            ch_evs.filter_channel(filter_channels)
        stout = mon_ch_addr.rawoutput_from_chevents(
            ch_evs,
            normalize=False,
            filter_duplicates=False
            )
        self.monitors.populate_monitors(chstlist=stout)
        return stout

    def run(self, *args, **kwargs):
        '''
        Prepares and stimulates.
        *args* and *kwargs* are keyword arguments passed to Communicator.run()
        '''
        self.prepare()
        if not self.offline:
            return self.stimulate(*args, **kwargs)
        else:
            return np.zeros([0,2],dtype='uint32')

    def stimulate_raw(self, raw_stim, **kwargs):
        '''
        Calls communicator.run without pre- or post-processing
        *raw_stim*: a numpy array in (addr, time) format. Type should be uint32 (shape = (-1,2)).
        Useful for debugging purposes. Format of raw_stim corresponds to the one defined in ComAPI:

        '''
        return self.communicator.run(raw_stim, **kwargs)

    def stimulate(self, stim=None, **kwargs):
        '''
        Run without preparing.
        Pre-processes, runs communicator, and post-processes.
        *kwargs* are keyword arguments passed to self.communicator.run()
        Returns a Stas.RawOutput object and populates monitors. The latter is the preferred way of reading data out.
        '''
        stim_evs = self._pre_process(stim)
        #run rec for run and record (consider using run for recording TODO)
        evs = self.communicator.run_rec(stim_evs, **kwargs)
        return self._post_process(evs, self.monitors.channels)





#-----------------------------------------------------------------------------
# Purpose:
#
# Author: <authors name>
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

#Path for recording experimental data
from .ComAPI import *
import collections 

class BatchCommunicator(BatchCommunicatorBase):
    def __init__(self):
        '''
        The BatchCommunicator API defined by the NCS tools.

        Usage:
        >>> c = Communicator(host = 'localhost', devnum = 0)
        >>> c.run()
        '''

#        Inputs:
# *host:* the hostname of the computer where the hardware is attached (default:
# localhost).
        ResourceManagerBase.__init__(self)
        RecordableCommunicatorBase.__init__(self)
        self.t_last = 0.
        self.queue = collections.deque()

    def run(self, stimulus=None, duration=None, context_manager=None):
        '''
        Loopback API simply puts an event packet in the API, and reads it back
        '''
        stimulus[:,1] = stimulus[:,1].cumsum()+self.t_last
        t_last = stimulus[-1,1]
        self.queue.appendleft(stimulus)
        return self.queue.pop()


class ContinuousCommunicator(BatchCommunicator):

    def mon(self, duration=None):
        #IMPLEMENT
        '''
        Returns one event packet from the queue independent of the durection
        '''
        return self.queue.pop()
        

    def stim(self, stimulus, duration=None, context_manager=None, **stim_kwargs):
        #IMPLEMENT
        '''
        Adds an event packet to the queue, independent of duration and reads it.
        '''
        stimulus[:,1] = stimulus[:,1].cumsum()+ self.t_last
        t_last = stimulus[-1,1]
        self.queue.appendleft(stimulus)
        return self.mon()

    def run(self, stimulus=None, duration=None, context_manager=None):
        '''
        Loopback API simply puts an event packet in the API, and reads it back
        '''
        return self.stim(stimulus)



        



Communicator = ContinuousCommunicator



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: <authors name>
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------
# -*- coding: utf-8 -*-


from .chip_v2 import Chip, NeuronBlock, NeuroChip, Block
from . import experimentTools as et
from .group import AddrGroup, AddrGroupBase
from .population import Population
from .connection import Connection, PConnection
from .mapping import Mapping, PMapping
from .neurosetup import NeuroSetup
from .monitors import Monitors, SpikeMonitor
from .api import ComAPI
from .api import ConfAPI
from . import pyST

######################  Utility functions
from functools import wraps


class DocInherit(object):
    ## {{{ http://code.activestate.com/recipes/576862/ (r1)
    """
    Docstring inheriting method descriptor

    The class itself is also used as a decorator

    doc_inherit decorator

    Usage:

    class Foo(object):
        def foo(self):
            "Frobber"
            pass

    class Bar(Foo):
        @doc_inherit
        def foo(self):
            pass

    Now, Bar.foo.__doc__ == Bar().foo.__doc__ == Foo.foo.__doc__ == "Frobber"
    """

    def __init__(self, mthd):
        self.mthd = mthd
        self.name = mthd.__name__

    def __get__(self, obj, cls):
        if obj:
            return self.get_with_inst(obj, cls)
        else:
            return self.get_no_inst(cls)

    def get_with_inst(self, obj, cls):

        overridden = getattr(super(cls, obj), self.name, None)

        @wraps(self.mthd, assigned=('__name__', '__module__'))
        def f(*args, **kwargs):
            return self.mthd(obj, *args, **kwargs)

        return self.use_parent_doc(f, overridden)

    def get_no_inst(self, cls):

        for parent in cls.__mro__[1:]:
            overridden = getattr(parent, self.name, None)
            if overridden:
                break

        @wraps(self.mthd, assigned=('__name__', '__module__'))
        def f(*args, **kwargs):
            return self.mthd(*args, **kwargs)

        return self.use_parent_doc(f, overridden)

    def use_parent_doc(self, func, source):
        if source is None:
            raise NameError("Can't find {o} in parents".format(self.name))
        func.__doc__ = source.__doc__
        return func

doc_inherit = DocInherit





#-----------------------------------------------------------------------------
# Purpose:
#
# Author: <authors name>
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

#Path for recording experimental data
import numpy, getpass, warnings
EMPTY_RAW = numpy.zeros([0,2], dtype='uint32')
USER = getpass.getuser()

class ResourceManagerBase(object):
    '''
    The ResourceManagerBase class is a base class for opening, closing a resource. It is used as parent classes for the configuration, communication and mapping APIs.
    The init function takes no arguments by default
    '''
    def __init__(self):
        self._neurosetup = None
        self._neurosetup_registered = False
        self._isopen = False

    def open(self):
        '''
        Opens resource
        '''
        self._isopen = True

    def close(self):
        '''
        Closes resource
        '''
        self._isopen = False

    @property
    def isopen(self):
        return self._isopen

    def __del__(self):
        if self.isopen:
            self.close()

    @property
    def neurosetup(self):
        if not self._neurosetup_registered:
            warnings.warn('NeuroSetup has not been registered')
            return None
        else:
            return self._neurosetup

    def register_neurosetup(self, neurosetup):
        '''
        Provides a link to the Neurosetup. This is useful for complex parameter
        configuration protocols requiring the sequencing and monitoring of
        address-events
        '''
        self._neurosetup_registered = True
        self._neurosetup = neurosetup

class RecordableCommunicatorBase(object):
    REC_PATH = '/tmp/exp_rec_' + USER
    REC_FN_SEQ = 'seq'
    REC_FN_MON = 'mon'
    REC_HEADER_SEQ = '# File format raw address, ISI (us)'
    REC_HEADER_MON = '# File format raw address, timestamp (us)'

    def __init__(self):
        self._rec_fns = []
        self._run_id = 0
        self.__reset()
        self.del_all()

    def get_exp_rec(self):
        '''
        Returns the filename where the raw stimated and monitored events are saved.
        The filenames for all experiments are returned, and run_id is reset

        Output: list of filenames containing the experiment records (stimulated and monitored events). Additionally, the experiment number is reset.
        
        '''        
        #CONVIENIENCE FUNCTION, IMPLEMENTATION NOT REQURIED
        import copy
        rec_fns = copy.copy(self._rec_fns)
        self.__reset()
        return rec_fns

    def run_rec(self, stimulus = None, *args, **kwargs):
        #CONVIENIENCE FUNCTION, IMPLEMENTATION NOT REQURIED
        '''
        Stimulate the neuromorphic hardware with event stream *stimulus* and monitor it for *duration* ms.
        In addition to the run() function, this function records the experimental data in /tmp/.
        The resulting files can be obtained with get_exp_record
        This function should be overridden to avoid redundant events file creation.

        Input: (see also run() for more information)
        *stimulus*: a numpy array in (addr, time) format. Type should be uint32 (shape = (-1,2)).
        *duration*: monitor duration.
        *context_manager*: context manager used to wrap the stimulate function. Useful for syncing with external process.
        Output: a numpy array in addr, time format. Type is uint32 (shape = (-1,2)).

        Usage:
        >>> events = np.array([[0, 1],[100,200]], dtype='uint32') #addr, time format
        >>> run_record(events, 1500)
        '''
        #CONVIENIENCE FUNCTION, IMPLEMENTATION NOT REQUIRED
        stim_fn, mon_fn = self.__gen_rec_fns()
        #Save stim before in case something goes wrong        
        self.__save_rec_file(stimulus, stim_fn)
        mon_evs = self.run(stimulus = stimulus, *args, **kwargs)
        self.__save_rec_file(mon_evs, mon_fn)

        return mon_evs

    def __save_rec_file(self, ev_array, filename):
        '''
        Save data using np.save, and adds filename in self._record_fns
        '''
        self._rec_fns.append(filename+'.npy')
        numpy.save(filename, ev_array)

        self._run_id += 1

    def __gen_rec_fns(self):
        '''
        Generate filenames for recording
        '''
        #Get username


        import time
        N = self._run_id
        current_time_str = time.strftime("__" + "%d-%m-%Y-%H:%M:%S", time.localtime())
        filename = self.REC_PATH + '_{2}_{0}__run{1}'
        stim_fn = filename.format(current_time_str, N, self.REC_FN_SEQ)
        mon_fn = filename.format(current_time_str, N, self.REC_FN_MON )
        return stim_fn, mon_fn

    def __reset(self):
        self._run_id = 0
        self._rec_fns = []
        return None

    def reset(self):
        #CONVIENIENCE FUNCTION, IMPLEMENTATION NOT REQURIED
        pass
    
    def del_all(self):
        import glob, os
        fn_to_del = glob.glob(self.REC_PATH+'*')
        for f in fn_to_del:
            os.remove(f)

    def __del__(self):
        self.del_all()

    def run(self, *args, **kwargs):
        return EMPTY_RAW






# -*- coding: utf-8 -*-
from setuptools import setup

setup(name='pyNCSre',
	version='20180119',
	description='Python Neurormophic Chips and Systems, reworked',
	author='Emre Neftci',
	author_email='eneftci@uci.edu',
	url='https://github.com/nmi-lab/pyNCS',
        download_url='https://github.com/nmi-lab/pyNCS/tarball/stable_20170814',
	packages = ['pyNCSre', 'pyNCSre.pyST', 'pyNCSre.api'],
        package_dir={'' : 'src'},
	package_data={'pyNCSre' : ['data/*.dtd',
                             'data/chipfiles/*.csv',
                             'data/chipfiles/*.nhml',
                             'data/chipfiles/*.xml']},
        include_package_data=True,
        install_requires=['numpy',
                          'lxml',
                          'matplotlib'],
     )



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Fabio Stefanini
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

import itertools
from warnings import warn
from .pyST import *

class AddrGroupBase(object):
    def __init__(self, name, description=None):
        if description:
            self.description = description
        else:
            self.description = 'group'
        self.name = name  # a short name
        self._channel = None
        self.chipid = ''
        self.addr = None
        self._paddr = None
        self._laddr = None
        self._channel = None
        self.grouptype = ''
        self.dtype = None
    
    def __deepcopy__(self, memo):
        return self.__copy__()

    def __copy__(self):
        '''
        Return an exact copy of the group.
        '''
        g = AddrGroup(self.name, self.description)
        g._channel = self._channel
        g.chipid = self.chipid
        g.addr = self.addr.copy()
        g.setup = self.setup
        g.addrspec = self.addrspec
        g.grouptype = self.grouptype
        g.dtype = self.dtype
        g.repopulate()
        return g

    @property
    def channel(self):
        return self._channel
    
    @channel.setter
    def channel(self, ch):
        self._channel = ch 
    
    @property
    def laddr(self):
        return self._laddr
    
    @property
    def paddr(self):
        return self._paddr   
    
    @laddr.setter
    def laddr(self, laddr):
        self._laddr = laddr
    
    @paddr.setter
    def paddr(self, paddr):
        self._paddr = paddr    
 
     
    def __len__(self):
        """
        x.__len__() ==> len(x)
        """
        return len(self.addr)

    def __eq__(self, other):
        '''
        Compares two address groups
        '''
        return (self.paddr == other.paddr).all()

    def __getitem__(self, i):
        """
        x.__getitem__(i) ==> x[i]
        """
        g = self.__copy__()
        if isinstance(i, slice):
            g.addr = self.addr[i].copy()
            if self._paddr != None:
                g._paddr = self._paddr[i]
            if self._laddr != None:
                g._laddr = self._laddr[i]
        else:
            if hasattr(i, '__len__'):
                g.addr = np.array(self.addr[i])
                if self._paddr != None:
                    g._paddr = np.array(self._paddr[i], dtype='uint32')
                if self._laddr != None:
                    g._laddr = np.array(self._laddr[i], dtype='float')
            else:
                g.addr = np.array([self.addr[i]], dtype=self.dtype)
                if self._paddr != None:
                    g._paddr = np.array([self._paddr[i]], dtype='uint32')
                if self._laddr != None:
                    g._laddr = np.array([self._laddr[i]], dtype='float')
        return g

    def __getslice__(self, i, j):
        """
        x.__getslice__(i, j) ==> x[i:j]
        """
        g = self.__copy__()
        g.addr = self.addr[i:j]
        if self._paddr != None:
            g._paddr = self._paddr[i:j]
        if self._laddr != None:
            g._laddr = self._laddr[i:j]
        return g


    def sort(self, order=None):
        """
        Sort all the addresses with the given order.
        If the order is None it orders by the last column of the address (most probably the column of synapses).
        Otherwise order must be a list of the fields in self.addr. 
        This is very useful for sorting 2D+ addresses.        
        """
        if order is None:
            idx = np.argsort(self.addr)
        else:
            idx = np.lexsort(list(map(self.addr.__getitem__, order)))
        # Just to ensure the addresses are generated.
        self.laddr; self.paddr
        # Sort the addresses accordingly.
        self.addr = self.addr[idx]
        self._laddr = self._laddr[idx]
        self._paddr = self._paddr[idx]

    def add(self, addresses):
        """
        Add a list of addresses to a group.
        """
        if self.grouptype is '':
            raise Exception("Group has never been populated before.")
        
        if len(addresses) == 0:
            return

        # Check for data type
        if isinstance(addresses, np.ndarray):
            if np.issubdtype(addresses.dtype, self.dtype):
                addr = addresses
            else:
                # WARNING: This data type conversion can introduce errors for
                # complex data types.
                addr = addresses.astype('uint32').view(self.dtype)
                addr = addr.T[0]
        else:
            addr = np.array(addresses, dtype='uint32').view(self.dtype)
            addr = addr.T[0]

        if len(self.addr) > 0:
            self.addr = np.concatenate([self.addr, addr])
        else:
            self.addr = addr


    def is_empty(self):
        return len(self) == 0

    # TODO: convert spiketrains methods to spiketrains(fashion='poisson') style
class AddrGroup(AddrGroupBase):
    #TODO: merge function
    """
    A AddrGroup is a group of chip addresses. They can be of grouptype 'in' or
    of grouptype 'out'. A description is needed for populating the group.

    Ex.:
        setup = pyNCS.Setup('setuptype.xml')
        setup.load('setup')
        setup.apply()

        # output from the retina
        group1 = AddrGroup('retina output')
        group1.populate_rectangle(
            setup, 'retina', 'out', [0,0,0], [64,64,0], n = 28*28 )

        # input from the ifslwta_0 chip
        group2 = AddrGroup('A1 cortex learning input')
        group2.populate_rectangle(setup, 'ifslwta_0', 'in', [0,4], [28,32])

        # input from the sequencer
        group3 = AddrGroup('A1 cortex excitatory input')
        group3.populate_rectangle(setup, 'sequencer', 'in', [0,2], [28,2])

        # input from the sequencer
        group4 = AddrGroup('A1 cortex excitatory inhibitory')
        group4.populate_rectangle(setup, 'ifslwta_0', 'in', [0,2], [28,2])

        group5 = AddrGroup('56 trains from sequencer')
        group4.populate_by_number(setup, 'seq', 'out', 56)

        # you want to treat group3 and group4 as a whole
        group4.join_with(group3)

        mapping = []
        mapping = group1.connect_one2one(group2)
             # the connect functions will maybe go in some Mapping module
        # ... and so on and so forth ...

        setMappings(mapping)

    """



    @property
    def ch_addr(self):
        if self.grouptype == 'in':
            return self.setup.seq
        elif self.grouptype == 'out':
            return self.setup.mon
        elif self.grouptype == '':
            warn('Grouptype has not been set')
            return None
        elif self.grouptype == '':
            raise ValueError('Grouptype should be None, "in" or "out", not {0}'.
                format(self.grouptype))

    @property
    def channel(self):
        if self._channel == None:
            # if it is None generate the necesary addresses
            if self.grouptype == 'in':
                self._channel =\
                self.setup.aerSlot[self.setup.chipslots[self.chipid]]['seqIn'][0]
            elif self.grouptype == 'out':
                self._channel =\
                self.setup.aerSlot[self.setup.chipslots[self.chipid]]['monOut'][0]

        return self._channel



    def repopulate(self, setup = None):
        """
        Repopulates laddr and paddr with respect to addr
        """
        if setup is not None:
            self.setup = setup.chaddrspecs
        if self.grouptype is '':
            raise Exception("Group has never been populated before.")
        self._channel = None
        self._laddr = None
        self._paddr = None
        self.addr = self.addr.copy()


#    def remove(self, address):
#        """
#        Remove list of addresses to a group.
#        """
#
#        if self.grouptype is '':
#            raise Exception("Group has never been populated before.")
#
#        raise NotImplementedError

    def __populate__(self, setup, chipid, grouptype, addresses=[]):

        self.chipid = chipid
        self.grouptype = grouptype
        self.setup = setup.chaddrspecs

        if self.grouptype is 'in':
            self.addrspec = setup.seq[self.channel]
        elif self.grouptype is 'out':
            self.addrspec = setup.mon[self.channel]
        else:
            self.addrspec = None

        self.dtype = self._get_dtype(setup, chipid, grouptype)
        self.addr = np.array([], dtype=self.dtype)
        self._paddr = None  # np.array([],dtype='uint32')
        self._laddr = None  # np.array([],dtype='float')
        self.add(addresses)
    
    def _get_dtype(self, setup, chipid, grouptype):
        '''
        Returns the data type of AddrGroup.addr variable, ie. the
        format of human readable addresses.
        '''
        if grouptype == 'in':
            flds = setup.get_chip_aerin(chipid).addrDict
        elif grouptype == 'out':
            flds = setup.get_chip_aerout(chipid).addrDict
        else:
            raise ValueError('Grouptype should be None, "in" or "out", not {0}'.
                format(grouptype))
        dtp = [None for i in range(len(flds))]
        for k,v in list(flds.items()):
            dtp[v] = (k, 'uint32')
        return np.dtype(dtp)


    def populate_line(self, setup, chipid, grouptype, addresses):
        self.__populate__(setup, chipid, grouptype, addresses)

    def populate_rectangle(self, setup, chipid, grouptype, p1, p2, z=None, n=None):
        """
        Populate with all the neurons within the rectangle's given coordinates. If n is given,
        populate with the given number of neurons instead of all neurons in the rectangle.
        z defines the 3d coordinate, e.g. retina polarity or synapse in the 2d chip.
        """

        if n:
            shaperatio = (1. * p2[1] - p1[1]) * 1. / (1. * p2[0] - p1[0])
            nx, ny = int(
                np.sqrt(1. * n / shaperatio)), int(np.sqrt(1. * shaperatio * n))
            addrx = np.linspace(p1[0], p2[0], nx)
            addry = np.linspace(p1[1], p2[1], ny)
        else:
            addrx = np.arange(p1[0], p2[0] + 1)
            addry = np.arange(p1[1], p2[1] + 1)

        if z == None:
            addresses = np.column_stack([addrx.repeat(len(
                addry)), np.array(addry.tolist() * len(addrx))])
        else:
            addresses = np.column_stack([addrx.repeat(len(addry)), np.array(addry.
                tolist() * len(addrx)), np.repeat(z, len(addrx) * len(addry))])

        self.__populate__(setup, chipid, grouptype, addresses)

    def populate_cuboid(self, setup, chipid, grouptype, p1, p2):
        '''
        Populate addresses within volume defined by the two opposite vertices
        in the N-dimensional space.
        setup: setup object
        chipid: ID of the chip to be populated on
        grouptype: IN or OUT
        p1: Lower Vertex (included) (vector of size n) defining the starting vertex
        p2: Upper Vertex (excluded) (vector of size n) defining the end vertex
        '''
        try:
            assert( len(p1)==len(p2) )
        except AssertionError as e:
            raise Exception("Dimensions of vertices do not match")
        try:
            assert( ((np.array(p2) - np.array(p1)) <= 0).sum() == 0 )
        except AssertionError as e:
            warn('Zero volume for given vertices')
        
        edges = []
        for i in range(len(p1)):
            edges.append(list(range(p1[i], p2[i])))
        addresses = itertools.product(*edges)
        self.__populate__(setup, chipid, grouptype, list(addresses))

#    def populate_by_number(self, setup, chipid, grouptype, n, dims):
#        """
#        Populate with the given number of neuron, doesn't matter the shape.
#        Needs the number of dimensions
#        WARNING: This should go in upper layer of abstraction.
#        TODO: check if this function is unused (Reason: contains reference errors)
#        """
#        max = np.prod([len(i) for i in dims])
#        if n > max:
#            raise Exception(
#                'Don\'t have enough space on this chip (max is %d).' % max)
#
#        lendims = [len(i) for i in dims]
#        nd = 0  # number of needed dimensions
#        div = 1
#        while True:
#            div = int(n / np.prod(lendims[0:nd + 1]))
#            if div > 0:
#                print "Dimension %d is not enough" % nd
#                nd += 1
#            else:
#                break
#
#        # unfold dimensions here
#        self.__unfold_dims__(dims)
#
#        dim = len(xdims)
#
#        n_filled_lines = int(n / dim)
#        filled_lines = np.arange(n_filled_lines)
#
#        n_return_line = n % dim
#        return_line = np.arange(n_return_line)
#
#        addrx = np.array(xdims)
#
#        addresses = np.column_stack([addrx.repeat(n_filled_lines),
#             np.array(filled_lines.tolist() * dim)])
#        addresses_ret = np.column_stack(
#            [return_line, np.repeat(filled_lines, n_return_line)])
#
#        addresses = np.concatenate([addresses, addresses_ret])
#
#        self.__populate__(setup, chipid, grouptype, addresses)


    def spiketrains(self, spikes=[], t_start=0, t_stop=None, dims=None):
        '''
        Constructs a stimulus with the desired SpikeList.
        spikes : list of tuples (id, time)
        id -> logical addresses of the group ie. AddrGroup.laddr
        All the arguments are the same as that of a pyST.SpikeList
        '''
        stStim = SpikeList(spikes=spikes, id_list=self.laddr,
             t_start=t_start, t_stop=t_stop, dims=dims)
        return {self.channel: stStim}

    def spiketrains_single(self, spike_time, t_after=100):
        """
        Creates a spiketrain with a single event with the given spike_time for all the addresses in the group.
        Inputs:
        *spike_time*: a float or a vector of floats indicating the spike times. If a vector is given, each entry is used once for each address in the logical addresses of the group (in the same order as self.laddr)
        *t_after* ms after the spike
        """
        s = np.empty([len(self), 2])
        s[:, 1] = spike_time
        s[:, 0] = self.laddr
        t_stop = max(s[:, 1]) + t_after
        spiketrain = self.spiketrains(s, t_start=0, t_stop=t_stop)
        return spiketrain

    def spiketrains_regular_gaussian(self, rate, offset=0., scale=5., t_start=0., duration=1000., channel=None):
        '''
        Returns a SpikeTrain whose spikes are regularly spaced, but jittered
        according to a Gaussian distribution around the spiking period, with
        the given rate (Hz) and stopping time t_stop (milliseconds).

        Note: t_start is always 0.0, thus all realizations are as if they
        spiked at t=0.0, though this spike is not included in the SpikeList.

        Inputs:
            rate    - the rate of the discharge (in Hz). Can be an iterable
            t_start - the beginning of the SpikeTrain (in ms)
            phase   - Use an offset vector to add an offset shift to the spike trains. Can be an iterable
            scale   - width of the Gaussian distribution placed at the regular
                      spike times, according to which the spike will be drawn (in ms)
            duration- The duration of the SpikeTrain (in ms)
            channel - Channel argument to enforce a particular channel
        '''
        stStim = SpikeList([], id_list=[])

        if not hasattr(rate, '__len__'):
            rate = rate * np.ones_like(self.laddr)
        if not hasattr(offset, '__len__'):
            offset = offset * np.ones_like(self.laddr)
        elif len(self.laddr) != len(rate):
            raise RuntimeError('Rate vector must be of the same length as the number of neurons in population: {0}'.format(len(self.addr)))

        if self.is_empty():
            print("AddrGroup is empty!")
            return stStim

        for i, id in enumerate(self.laddr):
            stStim[id] = STCreate.regular_gaussian_generator(rate[i],
                                                    phase=offset[i],
                                                    scale=scale,
                                                    t_start=t_start,
                                                    t_stop=t_start + duration)

        if channel is None:
            channel = self.channel

        return {channel: stStim}

    def spiketrains_regular(self, rate, offset=0., t_start=0., duration=1000.,
                            jitter=False, channel=None):
        """
        Create a regular spiketrain for each address on the group.
        You can use channel argument to enforce a particular channel.
        Use an offset vector to add an offset shift to the trains (must be a
        vector).

        Inputs:
            rate    - the rate of the discharge (in Hz). Can be an iterable
            t_start - the beginning of the SpikeTrain (in ms)
            offet   - Offset the spiketrain by this number (in ms.). Can be an iterable.
            jitter  - whether the spiketrain should be jittered by an amount numpy.random.rand()/rate
            duration- The duration of the SpikeTrain (in ms)
            channel - Channel argument to enforce a particular channel
        """
        stStim = SpikeList([], id_list=[])

        if not hasattr(rate, '__len__'):
            rate = rate * np.ones_like(self.laddr)
        if not hasattr(offset, '__len__'):
            offset = offset * np.ones_like(self.laddr)
        elif len(self.laddr) != len(rate):
            raise RuntimeError('Rate vector must be of the same length as the number of neurons in population: {0}'.format(len(self.addr)))

        if self.is_empty():
            print("AddrGroup is empty!")
            return stStim

        for i, id in enumerate(self.laddr):
            stStim[id] = STCreate.regular_generator(rate[i],
                                                    phase=offset[i],
                                                    jitter=jitter,
                                                    t_start=t_start,
                                                    t_stop=t_start + duration)

        if channel is None:
            channel = self.channel

        return {channel: stStim}

    def spiketrains_poisson(self, rate, t_start=0., duration=1000.,
                            channel=None):
        """
        Create a poisson spiketrain for each address on the group.
        You can use channel argument to enforce a particular channel.
        """
        stStim = SpikeList([], id_list=[])

        if not hasattr(rate, '__len__'):
            rate = rate * np.ones_like(self.laddr)
        elif len(self.laddr) != len(rate):
            raise RuntimeError('Rate vector must be of the same length as the number of neurons in population: {0}'.format(len(self.addr)))

        if self.is_empty():
            print("AddrGroup is empty!")
            return stStim

        for i, id in enumerate(self.laddr):
            stStim[id] = STCreate.poisson_generator(
                rate[i], t_start=t_start, t_stop=t_start + duration)

        if channel is None:
            channel = self.channel

        return {channel: stStim}

    def spiketrains_inh_poisson(self, rate, t, channel=None, **kwargs):
        """
        Create inhomogeneous poisson spiketrains. Rate is a vector of rates.
        The first dimension of rates corresponds to the neurons
        The second dimension corresponds to the time bin. The length of this must be the same as t.
        In addition, the rates *must* end with 0 to mark the end of the last bin.
        keyword arguments kwargs are passed to the spiketrain generator.
        See also pyST.STCreate.inh_poisson_generator
        """
        return self.spiketrains_inh_generator(rate, t, channel, base_generator=STCreate.poisson_generator, **kwargs)

    def spiketrains_inh_generator(self, rate, t, channel=None, base_generator=None, **kwargs):
        """
        Create inhomogeneous spiketrains.
        The process is defined by the base_generator function. See pyST.STCreate for available functions.
        Rate is a vector of rates.
        The first dimension of rates corresponds to the neurons
        The second dimension corresponds to the time bin. The length of this must be the same as t.
        In addition, the rates *must* end with 0 to mark the end of the last bin.
        keyword arguments kwargs are passed to the spiketrain generator.
        See also pyST.STCreate.inh_poisson_generator
        Kewword arguments are passed to the spike generator. Default is regular.
        """
        if base_generator == None:
            base_generator = STCreate.regular_generator
        stStim = SpikeList([], id_list=[])

        if len(self.addr) != rate.shape[0]:
            raise RuntimeError('Rate vector must be of the same length as the number of neurons in population: {0}'.format(self.laddr))

        assert rate.shape[1] == len(
            t), "time vector must be compatible with time axis of rate"

        if self.is_empty():
            print("AddrGroup is empty!")
            return stStim
        for i, id in enumerate(self.laddr):
            stStim[id] = STCreate.inh_poisson_generator(rate[i],
                 t, t_stop=t[-1], base_generator=base_generator, **kwargs)

        if channel is None:
            channel = self.channel

        return {channel: stStim}

    def spiketrains_1D_bump(self, pos, width, ampl, t_start=0., duration=                            1000., channel=None):
        rate = self.__spiketrains_periodic_bump_vector(pos, width, ampl)

        return self.spiketrains_poisson(rate, t_start, duration,
                                        channel=channel)



    def __spiketrains_periodic_bump_vector(self, pos, width, ampl):
        from scipy.stats import norm
        N = len(self.addr) / 2
        rate = norm.pdf(np.arange(len(self.addr)), scale=width, loc=N)
        rate = np.roll(rate, pos - N)
        rate = rate / rate.max() * ampl
        return rate

    def getLaddr(self):
        '''
        Regenreates Locigal addresses from address list.
        '''
        if self._laddr != None:
            return self._laddr
        if len(self.addr) > 0:
            self._laddr = self.ch_addr[self.channel].addrLogicalConstruct(
                self.addr.view('uint32').reshape((-1, len(self.dtype))).T)
        else:
            self._laddr = np.array([], dtype='float')
        return self._laddr

    def setLaddr(self, ad):
        self._laddr = ad
        warn('Not advisable to modify AddrGroup.laddr. Change AddrGroup.addr and repopulate() instead')

    def getPaddr(self):
        '''
        Generates Physical addresses from the address list. To regenerate use self.repopulate
        '''
        if self._paddr != None:
            return self._paddr
        if len(self.addr) > 0:
            self._paddr = self.ch_addr.addrPhysicalConstruct(
                {self.channel: self.addr.view('uint32').reshape((-1,
                                                                 len(self.dtype))).T})
        else:
            self._paddr = np.array([], dtype='uint32')
        return self._paddr

    def setPaddr(self, ad):
        self._paddr = ad
        warn('Not advisable to modify AddrGroup.paddr. Change AddrGroup.addr and repopulate() instead')

    # Properties of AddrGroup
    laddr = property(getLaddr, setLaddr, doc='Logical addresses')
    paddr = property(getPaddr, setPaddr, doc='Physical addresses')
